package leila;

import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.ObjectOutputStream;
import java.util.*;

import javatools.administrative.Announce;
import javatools.administrative.Parameters;
import javatools.filehandlers.DeepFileSet;
import javatools.filehandlers.FileSet;
import javatools.parsers.NumberFormatter;
import leila.algorithm.OutputPairs;
import leila.algorithm.Test;
import leila.algorithm.Train;
import leila.classifiers.Classifier;
import leila.parsing.HTML2LGI;
import leila.parsing.LGParse;
import leila.relations.Relation;

/**
 This class is part of LEILA (http://mpii.de/yago-naga/leila).
 It is licensed under the Creative Commons Attribution License 
 (http://creativecommons.org/licenses/by/3.0) 
 by the author Fabian M. Suchanek (http://suchanek.name).

 Leila.java is the main user interface for Leila.
 */
public class Leila {

  /** Loads a given the relation */
  public static Relation relation(String rel) {
    Announce.doing("Loading relation",rel);
    Relation learnMe=null;
    try {
      learnMe = (Relation) Class.forName(rel).newInstance();
    } catch (Exception e) {
      try {
        learnMe = (Relation) Class.forName("leila.relation."+rel).newInstance();
      } catch (Exception ex) {
        Announce.failed();
        Announce.error("Could not find relation",rel);
      }
    }
    Announce.done();
    return(learnMe);
  }
  
  /** Returns a name for a trained classifier */
  public static String classifierFileName(Relation r, Classifier c, Collection<String> docs) {
    return(r+"_"+c+"_"+FileSet.newExtension(docs.iterator().next(), null)+".mdl");
  }
  
  /** Loads the classifier form leila.ini*/
  public static Classifier classifierFromIni() {
    String classifierName=Parameters.get("classifier", "leila.classifier.Counter");
    Announce.doing("Creating classifier");
    Classifier classifier=null;
    try {
      classifier = (Classifier) Class.forName(classifierName).newInstance();
    } catch (Exception e) {
      try {
      classifier = (Classifier) Class.forName("leila.classifier." + classifierName).newInstance();
      } catch (Exception ex) {
        Announce.failed();
        Announce.error("Could not find classifier",classifierName);
      }
    }
    Announce.done();
    return(classifier);
  }
  
  /** Runs HTML2LGI, LGParse, Train and Test in a sequence*/
  public static void main(String[] args) throws Exception {
    // Check for leila.ini
    try {
      Parameters.init("leila.ini");
    } catch (Exception e) {
      Announce.help(
          "   Welcome to Leila!\n", 
          "Leila requires the Link Grammar Parser.", 
          "Download this parser here:",
          "    http://www.link.cs.cmu.edu/link", 
          "Then create a file called \"leila.ini\" that contains",
          "    parserFolder = <LinkGrammarParserFolder>", 
          "    parserCommand = <command to parse the file %1 to the file %2>",
          "For example, for Windows, this could be", 
          "    parserFolder = c:\\Program Files\\link41b",
          "    parserCommand = cmd /c \"c:\\Program Files\\link41b\\link41b.exe\" < %1 > %2", 
          "For Unix, it could be",
          "    parserFolder = /scratch/user/link-4.1b", 
          "    parserCommand = sh -c /scratch/user/link-4.1b/parse < %1 > %2 ",
          "leila.ini should be placed in the current folder.", 
          "When Leila is run later, it will tell you other parameters",
          "it might require in leila.ini.");
    }
    // Check arguments
    if (args.length < 3) {
      Announce.help(
        "   Leila <relation> <output> <doc> ... > <verbose>\n",
        "Extracts pairs of a target relation from documents.",
        "<relation> - a class that implements leila.relations.Relation.java",
        "             (e.g. SimpleFunction)",
        "<output> - a file where the found pairs shall be stored",
        "<doc>... - corpus files (e.g. ../data/sample.htm),",
        "           may be given by wildcards, if enclosed by quotes,",
        "           will also be seeked in subfolders",
        "<verbose> - a file to which verbose output can be written (nul for none)");
    }
    // Check parsed
    if (!args[2].endsWith(".lgo")) {
      Announce.doing("Checking whether the corpus documents have been parsed");
      Announce.message("   (you can avoid this check by passing the LGO-files as parameters)");
      for (int i = 2; i < args.length; i++) {
        for (File f : new DeepFileSet(args[i])) {
          File lgo = FileSet.newExtension(f, "lgo");
          if (!lgo.exists()) {
            File lgi = FileSet.newExtension(f, "lgi");
            if (!lgi.exists()) HTML2LGI.main(f);
            if (!LGParse.main(f)) {
              Announce.error("Check whether the path the command for the");
              System.err.println("    Link Grammar Parser in leila.ini are correct.");
              System.err.println("    Enclose filenames by quotes.");
              System.exit(255);
            }
          }
        }
      }
      Announce.done();
    }
    Relation learnMe=relation(args[0]);
    Classifier classifier=classifierFromIni();
    List<String> files = Arrays.asList(args).subList(3, args.length);    
    String outputFile=args[1];
    
    // ------ Now let's get going    
    System.out.println("\n\nLeila Run");
    System.out.println("  Time: " + NumberFormatter.ISOtime());
    System.out.println("  Relation: " + learnMe.toString());
    System.out.println("  Classifier: " + classifier);
    System.out.println("  OutputFile: " + outputFile);
    System.out.println("  Documents: " + files + "\n");
    
    // Training
    Train.main(learnMe, classifier, files);        
    
    // Testing
    Announce.doneDoing("Opening output-file " + args[3]);
    OutputPairs p = new OutputPairs(new FileWriter(args[3]));
    Announce.done();
    Test.main(files, learnMe, classifier, p);
  }

}