package leila.algorithm;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.ObjectInputStream;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import javatools.administrative.Announce;
import javatools.administrative.Parameters;
import javatools.datatypes.CombinedIterator;
import javatools.datatypes.Visitor;
import javatools.filehandlers.DeepFileSet;
import javatools.parsers.NumberFormatter;
import leila.classifiers.Classifier;
import leila.patterns.LinearLinkageFile;
import leila.patterns.Link;
import leila.patterns.Linkage;
import leila.patterns.LinkageFile;
import leila.patterns.LinkagePattern;
import leila.patterns.LinkageWord;
import leila.patterns.StoredBridge;
import leila.patterns.Link.Direction;
import leila.relations.Relation;

/**
This class is part of LEILA (http://mpii.de/yago-naga/leila).
It is licensed under the Creative Commons Attribution License 
(http://creativecommons.org/licenses/by/3.0) 
by the author Fabian M. Suchanek (http://suchanek.name).
<P>

Test.java is a user-interface that applies learned patterns to a set of files.

*/

public class Test {

  /** Applies a visitor to the FeatureObjects of all pairs of interesting words in all
   *  sentences in a file */
  public static void apply(File inf, Visitor<LinkagePattern> visitor) throws Exception {
    for(Linkage linkage : new CombinedIterator<Linkage>(new LinkageFile(inf), new LinearLinkageFile(inf))) {
      for(LinkageWord w : linkage.linkageWords) {
        if(!w.is(LinkageWord.INTERESTING)) continue;
        runLinks(w,w,visitor,Link.Direction.RIGHT, new ArrayList<Link>(20),linkage.id);
        runLinks(w,w,visitor,Link.Direction.LEFT, new ArrayList<Link>(20),linkage.id);
      }
    }
  }

  /** Finds all FeatureObjects starting from LinkageWord start, calls visitor on each */
  private static void runLinks(LinkageWord start,LinkageWord current,
                               Visitor<LinkagePattern> v,Link.Direction d,
                               List<Link> links, int id) throws Exception {
    if(start!=current && current.is(LinkageWord.INTERESTING) && !start.word.equalsIgnoreCase(current.word))
      v.visit(new LinkagePattern(start,links,current,id));
    for(Link l : current.contexts) {
      if(l.direction!=d) continue;
      links.add(l);
      runLinks(start,l.destination,v,d,links,id);
      links.remove(links.size()-1);
    }
    return;
  }

  /** Runs through the sentences , applying the good patterns */
  public static void findNew(final File inf,final Relation learnMe,final Classifier memory,final OutputPairs newPairs) throws Exception {
    Announce.doing("Discovering output pairs in "+inf.getName());
    System.out.println("------ File "+inf+"\n");
    apply(inf, new Visitor<LinkagePattern>(){
       // Short term memory stores pairs found in
       // preceding FeatureObjects with same sourceId
       private Collection<Pair> stm=new ArrayList<Pair>();
       private int lastId=-1;
       public boolean visit(LinkagePattern v)  throws Exception {
         double confidence;
         // Clear reminders if we step into a new sentence
         if(lastId!=v.sourceId) stm.clear();
         if((confidence=memory.classify(v))>0.0) {
           Relation.WordPairType j=learnMe.assess(v.basis[0],v.basis[1]);
           if(j==Relation.WordPairType.CANDIDATE || j==Relation.WordPairType.EXAMPLE) {
             // Check whether we already know this pair
             Pair p=new Pair(v.basis[0].word,v.basis[1].word);
             if(!(lastId==v.sourceId && stm.contains(p))) {
                v.label=confidence;
                StoredBridge b=memory.bridgeMap.get(v.bridge);
                System.out.println("Found "+j+" "+v);
                newPairs.add(new Pair(v.basis[0],v.basis[1],confidence,inf.getName()+", Bridge: "+b+" "+j));
                stm.add(p);
             }
           }
         }
         lastId=v.sourceId;
         return(true);
       }
    });
    Announce.done();
  }

  /** Reports an error and quits */
  public static void error(String s) {
    System.err.println("done\nThe "+s+" was not specified");
    System.exit(2);
  }

  /** Applies the memory to the LGO-file(s) and prints all noun pairs found (call by Java)*/
  public static void main(Collection<String> files,Relation r, Classifier m, OutputPairs newPairs) throws Exception {
    System.out.println("----------------- Testing ----------------------\n");
    for(String fo : files) {
      for(File f : new DeepFileSet(fo)) {
        findNew(f,r,m,newPairs);
      }
    }
  }

  /** Applies the memory to the LGO-file and prints all noun pairs found (call by user)*/
  public static void main(String[] argv) throws Exception {
    if(argv.length<4) {
      System.err.println("\n\n   Test <rel> <model> <output> <LGOfile>... > <verbose>\n");
      System.err.println("Finds ouput pairs in a corpus.");
      System.err.println("<rel> - a class that implements leila.Relation.java");
      System.err.println("        (e.g. InstanceOf)");
      System.err.println("<model> - the file of the trained classifier");
      System.err.println("<output> - a file where the output pairs can be stored");
      System.err.println("<LGOfile>... - the parsed corpus documents,");
      System.err.println("               may contain wildcards if enclosed by quotes,");
      System.err.println("               will also be seeked in subfolders");
      System.err.println("<verbose> - file for verbose output (nul for none)");
      System.err.println("The first three arguments can be given in any order.");
      System.exit(63);
    }
    Parameters.init("leila.ini");
    Announce.doing("Mounting");
    Classifier memory=null;
    Writer out=null;
    Relation relation=null;
    System.out.println("Leila Test");
    System.out.println("  Time: "+NumberFormatter.ISOtime());
    for(String a : argv) {
      if(a.indexOf('.')==-1) {
        Announce.doneDoing("Loading relation "+a);
        try {
          relation=(Relation)Class.forName(a).newInstance();
        }
        catch(ClassNotFoundException e) {
          relation=(Relation)Class.forName("leila."+a).newInstance();          
        }
        System.out.println("  Relation: "+relation.getClass().getSimpleName());
        continue;
      }
      if(a.endsWith(".txt")) {
        Announce.doneDoing("Opening outputFile "+a);
        out=new FileWriter(a);
        System.out.println("  OutputFile: "+a);
        continue;
      }
      if(a.endsWith(".mem")) {
        Announce.doneDoing("Loading classifier "+a);
        ObjectInputStream dis=new ObjectInputStream(new FileInputStream(a));
        memory=(Classifier)dis.readObject();
        dis.close();
        System.out.println("  Classifier: "+memory.getClass().getSimpleName());
        System.out.println("  ClassifierFile: "+a);
        continue;
      }
    }
    Announce.done();
    if(memory==null) error("classifier");
    if(out==null) error("output-file");
    if(relation==null) error("relation");
    OutputPairs newPairs=new OutputPairs(out);
    ArrayList<String> files=new ArrayList<String>();
    for(int i=3;i<argv.length;i++) files.add(argv[i]);
    System.out.println("  Documents: "+files+"\n");
    main(files,relation,memory,newPairs);
  }
}