package leila.algorithm;
import java.io.File;
import java.io.FileOutputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;

import javatools.administrative.Announce;
import javatools.administrative.Parameters;
import javatools.datatypes.CombinedIterator;
import javatools.datatypes.Visitor;
import javatools.filehandlers.DeepFileSet;
import javatools.filehandlers.FileSet;
import javatools.parsers.NumberFormatter;
import leila.classifiers.Classifier;
import leila.patterns.LinearLinkageFile;
import leila.patterns.Linkage;
import leila.patterns.LinkageFile;
import leila.patterns.LinkagePattern;
import leila.relations.Relation;

/**
 This class is part of LEILA (http://mpii.de/~suchanek/leila).
 It is licensed under the Creative Commons Attribution License 
 (http://creativecommons.org/licenses/by/3.0) 
 by the author Fabian M. Suchanek (http://mpii.de/~suchanek).

Train.java trains a classifier for a given relation and a given corpus.
 */
public class Train {
  
  /** Trains on a LGO-file */
  public static void findPatternsWithExamples(final File document,Relation relation,Classifier classifier) throws /*lots of*/ Exception {
    Announce.doing("Finding examples in "+document.getName());
    System.out.println("----- File "+document+"\n");
    for(Linkage linkage : new CombinedIterator<Linkage>(new LinkageFile(document), new LinearLinkageFile(document))) {
      for(LinkagePattern fo : linkage.linkagePatterns(relation)) {
           System.out.println("Found example "+fo);
           classifier.add(fo);
      }
    }
    Announce.done();
  }

  /** Runs through the sentences again, finding counter examples */
  public static void validate(final File inf,final Relation learnMe,final Classifier memory) throws Exception {
    Announce.doing("Finding counterexamples in "+inf.getName());
    Test.apply(inf,new Visitor<LinkagePattern>(){
       public boolean visit(LinkagePattern v)  throws Exception {
         if(memory.classify(v)>0.0) {
           if(learnMe.assess(v.basis[0],v.basis[1])==Relation.WordPairType.COUNTEREXAMPLE) {
             System.out.println("Found counterexample "+v);
             v.label=-1.0;
             memory.add(v);
           }
         }
         return(true);
       }
    });
    Announce.done();
  }


  /** Trains a classifier and validates it from an LGO-file (call by Java)*/
  public static void main(Relation relation,Classifier classifier,Collection<String> documents) throws Exception {
    String classifierFile=leila.Leila.classifierFileName(relation, classifier, documents);
    System.out.println("----------------- Training ----------------------\n");
    System.out.println("Leila Training");
    System.out.println("  Time: "+NumberFormatter.ISOtime());
    System.out.println("  Relation: "+relation);
    System.out.println("  Classifier: "+classifier);
    System.out.println("  ClassifierFile: "+classifierFile);
    System.out.println("  Documents: "+documents+"\n");
    Announce.doing("Training");
    Announce.doing("Finding patterns with examples");
    for(String fo : documents) {
      for(File f : new DeepFileSet(fo)) {
        findPatternsWithExamples(f,relation,classifier);
      }
    }
    Announce.doneDoing("Finding patterns with counterexamples");
    if(relation.requiresCounterexamples()) {
      for(String fo : documents) {
        for(File f : new DeepFileSet(fo)) {
          validate(f,relation,classifier);
        }
      }
    }
    Announce.doneDoing("Training the classifier");
    classifier.train();
    Announce.doneDoing("Storing classifier to "+classifierFile);
    ObjectOutputStream oos=new ObjectOutputStream(new FileOutputStream(classifierFile));
    oos.writeObject(classifierFile);
    oos.close();
    Announce.done();
    Announce.done();
  }

  /** Trains a classifier and validates it from an LGO-file (call by User)*/
  public static void main(final String[] argv) throws Exception {
    if(argv.length<3) {
      System.err.println("\n\n   Train <relation> <model> <doc> ... > <verbose>\n");
      System.err.println("Trains a classifier for a relation.");
      System.err.println("<relation> - a class that implements leila.Relation.java");
      System.err.println("             (e.g. SimpleFunction)");
      System.err.println("<model> - a file where the trained classifier shall be stored");
      System.err.println("<doc>... - the parsed corpus documents,");
      System.err.println("           may contain wildcards if enclosed by quotes,");
      System.err.println("           will also be seeked in subfolders");
      System.err.println("<verbose> - file for verbose output (nul for none)");
    }
    Relation learnMe=leila.Leila.relation(argv[0]);
    Classifier classifier=leila.Leila.classifierFromIni();
    List<String> files = Arrays.asList(argv).subList(2, argv.length);    
    main(learnMe,classifier,files);
  }
}