package leila.classifiers;
import java.io.StringWriter;
import java.io.Writer;

import leila.patterns.LinkagePattern;
import leila.patterns.StoredBridge;
import leila.patterns.StoredLinkagePattern;
/**
This class is part of LEILA (http://mpii.de/yago-naga/leila).
It is licensed under the Creative Commons Attribution License 
(http://creativecommons.org/licenses/by/3.0) 
by the author Fabian M. Suchanek (http://suchanek.name).

<P>

Classifier.java represents a simple KNN-Classifier for a set of StoredFeatureObjects.
*/

public class KNNClassifier extends Classifier {

  /** All feature objects more similar than this value fall into the same equivalence class */
  public static double blackHole=0.5;

  /** Adds a new LinkagePattern */
  public void add(LinkagePattern o) throws Exception {
    StoredBridge b=bridgeMap.get(o.bridge);
    numFeatureObjects++;
    if(b==null) {
      b=new StoredBridge(o.bridge,o);
      bridgeMap.put(b.bridge,b);
      return;
    }
    StoredLinkagePattern maxf=null;
    double maxsim=0.0;
    for(StoredLinkagePattern s : b.users) {
      double sim=s.similarity(o);
      if(sim>maxsim) {
        maxsim=sim;
        maxf=s;
      }
    }
    if(maxsim>blackHole) {
      maxf.label+=maxsim*o.label;
      maxf.num++;
    } else {
      b.users.add(new StoredLinkagePattern(o));
    }
  }

  /** Returns the list of featureObjects, might be long, so use printTo instead! */
  public String toString() {
    StringWriter w=new StringWriter();
    try {
      printTo(w);
    } catch(Exception e){};
    return(w.getBuffer().toString());
  }

  /** Returns a similarity measure */
  public double classify(LinkagePattern o) throws Exception {
    StoredBridge b=bridgeMap.get(o.bridge);
    if(b==null) return(Double.NEGATIVE_INFINITY);
    if(!isContracted) return(1.0);
    if(b.users.size()==0) return(Double.NEGATIVE_INFINITY);
    StoredLinkagePattern best=null;
    double bestval=Double.NEGATIVE_INFINITY;
    for(StoredLinkagePattern f : b.users) {
      double sim=f.similarity(o);
      if(sim>bestval) {
        bestval=sim;
        best=f;
      }
    }
    return(bestval*best.label+500);
  }

  public void printTo(Writer o) throws Exception {
    o.write("************* kNN Classifier *************\n\n");
    for(StoredBridge b: bridgeMap.values()) {
       o.write("-----------------------------------------------------------------------\n");
       o.write("     Bridge "+b+"\n");
       o.write("-----------------------------------------------------------------------\n\n");
       for(StoredLinkagePattern s : b.users) o.write(s+"\n");
       o.write('\n');
    }
    o.flush();
  }

  /** Contracts the data */
  public void train() throws Exception {
    isContracted=true;
    for(StoredBridge b : bridgeMap.values()) {
      double sum=0.0;
      for(StoredLinkagePattern o : b.users) {
        sum+=o.label;
        o.label/=o.num;
      }
      b.sum=sum;
    }
  }
}