package leila.classifiers;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.InputStreamReader;
import java.io.Writer;

import javatools.administrative.Announce;
import javatools.administrative.D;
import javatools.administrative.Parameters;
import javatools.datatypes.SVMModel;
import javatools.datatypes.SparseVector;
import javatools.filehandlers.FileLines;
import javatools.parsers.NumberFormatter;
import leila.patterns.Link;
import leila.patterns.LinkagePattern;
import leila.patterns.LinkageWord;
import leila.patterns.StoredBridge;

/**
This class is part of LEILA (http://mpii.de/yago-naga/leila).
It is licensed under the Creative Commons Attribution License 
(http://creativecommons.org/licenses/by/3.0) 
by the author Fabian M. Suchanek (http://suchanek.name).

<P>

SVMClassifier.java implements a classifier by calling Thorsten Joachim's SVMLight
for each pattern. This class requires the following parameters in leila.ini:
<UL>
<LI> blackSheep - "on" to ignore bridges with only one positive occurence, "off" else</LI>
<LI> outFile - a filename that LEILA can use for temporary data</LI>
<LI> sortCommand - the command to sort the lines in the outFile.
     For Windows, this is "sort outFile /o outFile"</LI>
<LI> svmInFile - a filename that LEILA can use for temporary data</LI>
<LI> svmLightCommand - the command to launch the SVMLight learner on svmInFile, producing svmOutFile.<BR>
     For Windows, this is e.g. "c:\svmLight\svm_learn.exe <I>svmInFile svmOutFile</I>"</LI>
<LI> svmOutFile - a filename that LEILA can use for temporary data</LI>
</UL>
 */
public class SVMClassifier extends Classifier  {

  /** Holds a standard number formatter */
  private static final NumberFormatter formy=new NumberFormatter("#####");

  /** Holds the out-file */
  private transient BufferedWriter out=null;
  /** Bitwidth of connector in a feature vector */
  public static final int CONNECTORWIDTH=LinkageWord.MAXINT;
  /** Width of a basis in a feature vector */
  public static final int BASISWIDTH=CONNECTORWIDTH + Link.MAXINT*CONNECTORWIDTH;

  /** Creates an SVMClassifier */
  public SVMClassifier()  throws Exception {
    Parameters.ensureParameters(
      "blackSheep - \"on\" to ignore bridges with only one positive occurence, \"off\" else",
      "outFile - a filename that LEILA can use for temporary data",
      "sortCommand - the command to sort the lines in the outFile",
      "svmInFile - a filename that LEILA can use for temporary data",
      "svmLightCommand - the command to launch the SVMLight learner on svmInFile, producing svmOutFile",
      "svmOutFile - a filename that LEILA can use for temporary data"
    );
    out=new BufferedWriter(new FileWriter(Parameters.get("outFile")));
  }

  /** Creates a SparseVector from a LinkagePattern (after contraction)*/
  public SparseVector toSparseVector(LinkagePattern o) {
    // number of dimensions = size of both contexts + 2 Basiswords
    int numDim= (o.basis[0].contexts.size() +
                 o.basis[1].contexts.size())*
                 LinkageWord.MAXBITS;
    int[] dim=new int[numDim];
    int index=0;
    for(int i=0;i<2;i++) {
      for(Link l : o.basis[i].contexts) {
        index=l.destination.toArray(dim,index,i*BASISWIDTH+l.toInt()*CONNECTORWIDTH);
      }
    }
    return(new SparseVector(o.label,dim));
  }

  /** Classifies a LinkagePattern by the SVM */
  public double classify(LinkagePattern v2)  throws Exception {
    StoredBridge b=bridgeMap.get(v2.bridge);
    if(b==null) return(Double.NEGATIVE_INFINITY);
    if(!isContracted) return(1.0);
    double result=b.model.classify(toSparseVector(v2));
    // Results that are exactly 0.0 or -1.0 are always negative, upgrade the others
    if(result==0.0 || result==-1.0) return(Double.NEGATIVE_INFINITY);
    else return(result+5);
  }

  /** Executes a command */
  public static void execute(String cmd) throws Exception {
    Process p=Runtime.getRuntime().exec(cmd);
    BufferedReader bri=new BufferedReader(new InputStreamReader(p.getInputStream ()));
    BufferedReader bre=new BufferedReader(new InputStreamReader(p.getErrorStream ()));
    String s;
    while(null!=(s=bri.readLine()) || null!=(s=bre.readLine())) {
      System.out.println(s);
    }
    p.waitFor();
  }

  /** To be called after all Patterns have been added */
  public void train()  throws Exception{
    boolean blackSheep=Parameters.getBoolean("blackSheep");
    Announce.doing("Closing outfile");
    D.writeln(out,"lastooooo");
    out.close();
    // Sort the FeatureVectors by their bridges
    Announce.doneDoing("Sorting outfile");    
    execute(Parameters.get("sortCommand"));
    Announce.done();
    // Distribute the FeatureVectors to files, cut out black sheep,
    // eliminate duplicate word pairs
    // Call SVMLight
    String currentId="";
    String previous="";
    int pos=0;
    int neg=0;
    Announce.doing("Writing data for bridge 00000");
    out=new BufferedWriter(new FileWriter(Parameters.get("svmInFile")));
    for(String s : new FileLines(Parameters.get("outFile"))) {
      // Ignore identical lines
      if(previous.equals(s.substring(0,s.indexOf('')))) continue;
      previous=s.substring(0,s.indexOf(''));
      if(currentId.length()==0) currentId=s.substring(0,5);
      if(!currentId.equals(s.substring(0,5))) {
        out.close();
        Announce.doneDoing("Calling SVMLight on bridge "+currentId);
        // In case of black sheep
        if(blackSheep && pos==1 && (neg>10 || neg==0)) {
          out=new BufferedWriter(new FileWriter(Parameters.get("svmInFile")));
          D.writeln(out,"-1 1:0 # Black sheep");
          out.close();
        };
        // Call SVM
        System.out.println("------ Calling SVM for bridge "+currentId);
        execute(Parameters.get("svmLightCommand"));
        Announce.doneDoing("Reading model of bridge "+currentId);
        for(StoredBridge b : bridgeMap.values()) {
         if(b.id==Integer.parseInt(currentId)) {
           b.model=new SVMModel(Parameters.get("svmOutFile"));
           break;
         }
        }
        pos=neg=0;
        currentId=s.substring(0,5);
        if(currentId.charAt(0)!='l') {
          Announce.doneDoing("Writing data for bridge "+currentId);
          out=new BufferedWriter(new FileWriter(Parameters.get("svmInFile")));
        } else continue;
      }
      s=s.substring(s.indexOf('')+1);
      D.writeln(out,s);
      if(s.charAt(0)=='-') neg++;
      else pos++;
    }
    out.close();
    Announce.done();
    isContracted=true;
  }

  /** Adds a new LinkagePattern */
  public void add(LinkagePattern o)  throws Exception {
    StoredBridge b=bridgeMap.get(o.bridge);
    if(b==null) {
      b=new StoredBridge(o.bridge,o);
      bridgeMap.put(b.bridge,b);
    }
    out.write(formy.format(b.id)+o.basis[0].word+' '+o.basis[1].word+''+toSparseVector(o));
    out.write(" # "+o.bridge+' '+o.basis[0].word+' '+o.basis[1].word);
    out.write("\n");
    numFeatureObjects++;
  }

  public void printTo(Writer o) throws Exception {
    o.write("************* SVM Classifier *************\n\n");
    for(StoredBridge b: bridgeMap.values()) {
       o.write("-----------------------------------------------------------------------\n");
       o.write("     Bridge "+b+"\n");
       o.write("-----------------------------------------------------------------------\n\n");
       o.write(b.model.toString());
       o.write('\n');
    }
    o.flush();
  }

}