package leila.evaluate;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.geom.Ellipse2D;
import java.io.File;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javatools.administrative.Announce;
import javatools.filehandlers.DeepFileSet;
import javatools.filehandlers.FileLines;
import javatools.filehandlers.FileSet;
import javatools.parsers.Char;
import javatools.parsers.NumberFormatter;

import javax.swing.JApplet;
import javax.swing.JFrame;

import leila.parsing.HTMLReader;
import leila.relations.InstanceOf;

/**
This class is part of LEILA (http://mpii.de/yago-naga/leila).
It is licensed under the Creative Commons Attribution License 
(http://creativecommons.org/licenses/by/3.0) 
by the author Fabian M. Suchanek (http://suchanek.name).

<P>

Evaluate.java compares the output pairs of a system (e.g. LEILA) to ideal pairs.
The ideal pairs for file X.lgo should be in a file X.xxx (where xxx is an
arbitrary but fixed extension). This file should contain lines of the following form:
<PRE>
% Optional comments
first word / second word
first word / second word
...
</PRE>
This is the output produced by HandTag.java.
*/
public class Evaluate extends JApplet {

  /** A list of supported systems */
  public enum IESystem {LEILA,TEXT2ONTO,TEXTTOONTO,SNOWBALL,CV,CVSIMPLE};

  /** Holds a WordNet for approximate comparisons */
  public static InstanceOf wordNet;

  /** Holds an array of precision/recall-values [p1,r1,p2,r2,...]*/
  public double[] array;

  /** Constructor from an array of precision/recall-values */
  public Evaluate(double[] a) {
    array=a;
    setPreferredSize(new Dimension(500, 500));
  }

  /** Draws the precision/recall-dots*/
  public void paint(Graphics g) {
    Graphics2D g2=(Graphics2D) g;
    g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
          RenderingHints.VALUE_ANTIALIAS_ON);
    Dimension d=getSize();
    for(int i=0; i<array.length-1; i+=2) {
      g2.draw(new Ellipse2D.Double(array[i+1]*d.width, (1-array[i])*(d.height-10),5,5));
    }
  }

  /** Does nothing */
  public void init() {
  }

  /** Holds the files for which ideal pairs exist */
  public static Set<String> files=new TreeSet<String>();

  /** Holds ideal pairs */
  public static java.util.List<EvaluatePair> idealPairs=new ArrayList<EvaluatePair>();

  /** Holds output pairs */
  public static java.util.List<EvaluatePair> outputPairs=new ArrayList<EvaluatePair>();

  /** Adds the ideal pairs from a given file. Ideal pairs are given by the form
   * <PRE>
   * % Optional comments
   * first word / second word
   * first word / second word
   * ...
   * </PRE>
   * */
  public static void addIdeal(File f) throws Exception {
    files.add(FileSet.newExtension(f.getName(),""));
    for(String s : new FileLines(f)) {
      if(s.length()==0||s.charAt(0)=='%') continue;
      int i=s.indexOf('/');
      if(i==-1) continue;
      EvaluatePair p=new EvaluatePair(s.substring(0, i-1), s.substring(i+1), 1.0,FileSet.newExtension(f.getName(),""));
      idealPairs.add(p);
    }
  }

  /** Pattern of LEILA output pairs */
  public static final Pattern leilaPattern=Pattern
        .compile("(\\d{3}\\.\\d{10}) ([^ ]*) ([^ ]*) # ([^,]*), Bridge: #(\\d+).*(T|L)E");
  /** Pattern of textToOnto output pairs */
  public static final Pattern textToOntoPattern=Pattern
       .compile(".+:.+\\((.+),(.+)\\).*");
  /** Pattern of text2onto output pairs */
  public static final Pattern text2OntoPattern=Pattern
       .compile("instance-of\\( ([^,]+), ([^\\)]+) \\)");
  /** Pattern of CV output pairs */
  public static final Pattern cvPattern=Pattern
       .compile("([^:]+): ([^\\(]+) \\(([^\\)]+)\\)");
  /** Pattern of CV output pairs in their simple form */
  public static final Pattern cvSimplePattern=Pattern
       .compile("([^:]+):(.+)");

  /** Adds the outputPairs pairs from a given result file.
   * Skips index.lgo.*/
  public static void addOutput(File f, IESystem system) throws Exception {
    if(f.getName().equals("index.lgo")) return;
    switch(system) {
      case SNOWBALL:
        HTMLReader in=new HTMLReader(new FileReader(f));
        while(true) {
          String company=in.readTaggedText("FieldValue");
          if(company==null) break;
          company=company.replace('+', ' ')
                  .trim()
                  .replace(' ','_')
                  .replaceAll("_(\\W)", "$1");
          company=Char.decodeAndNormalize(company);
          String hq=in.readTaggedText("FieldValue").replace('+', ' ').trim().replace(' ','_');
          double rank=Double.parseDouble(in.readTaggedText("TupleRank"));
          String file=in.readTaggedText("SupportingDocId");
          file=FileSet.newExtension(file.substring(file.lastIndexOf('\\')+1),"");
          if(!files.contains(file)) continue;
          EvaluatePair p=new EvaluatePair(company,hq,rank,FileSet.newExtension(file,""));
          outputPairs.add(p);
        }
        in.close();
        break;
      case TEXT2ONTO:
        for(String s : new FileLines(f,"Reading ouput pairs")) {
          Matcher m=text2OntoPattern.matcher(s);
          while(m.find()) {
            outputPairs.add(new EvaluatePair(m.group(1),m.group(2),1.0,""));
          }
        }
        break;
      case TEXTTOONTO:
        for(String s : new FileLines(f,"Reading ouput pairs")) {
          Matcher m=textToOntoPattern.matcher(s);
          if(!m.matches()) continue;
          EvaluatePair p;
          if(s.startsWith("hearst")) p=new EvaluatePair(m.group(1), m.group(2), 1.0, "");
          else p=new EvaluatePair(m.group(2), m.group(1), 1.0, "");
          outputPairs.add(p);
        }
        break;
      case CV:
        String current="";
        String best="";
        double bestscore=-100.0;
        for(String s : new FileLines(f,"Reading ouput pairs")) {
          Matcher m=cvPattern.matcher(s);
          if(!m.matches()) continue;
          double c=0;
          try {
            c=Double.parseDouble(m.group(3));
          }
          catch(Exception e) {};
          if(!current.equals(m.group(1))) {
            if(current.length()!=0) {
              EvaluatePair p=new EvaluatePair(current,best,bestscore,"");
              outputPairs.add(p);
            }
            current=m.group(1);
            bestscore=c;
            best=m.group(2);
            continue;
          }
          if(bestscore<c) {
            best=m.group(2);
            bestscore=c;
          }
        }
        outputPairs.add(new EvaluatePair(current,best,bestscore,""));
        break;
      case CVSIMPLE:
        for(String s : new FileLines(f,"Reading ouput pairs")) {
          Matcher m=cvSimplePattern.matcher(s);
          if(!m.matches()) continue;
          outputPairs.add(new EvaluatePair(m.group(1), m.group(2), 1.0, ""));
        }
        break;
      case LEILA:
        for(String s : new FileLines(f,"Reading ouput pairs")) {
          Matcher m=leilaPattern.matcher(s);
          if(!m.matches()) continue;
          if(files.size()!=1 && !files.contains(FileSet.newExtension(m.group(4),""))) continue;
          double c=0;
          int bridgeId=-1;
          try {
            c=Double.parseDouble(m.group(1));
            bridgeId=Integer.parseInt(m.group(5));
          }
          catch(Exception e) {};
          EvaluatePair p=new EvaluatePair(m.group(2), m.group(3), c, FileSet.newExtension(m.group(4),""));
          p.bridgeId=bridgeId;
          if("L".equals(m.group(6))) p.example=true;
          outputPairs.add(p);
        }
        break;
    }
    files=null; // We don't need it any longer, free the memory
  }

  /** Calculates precision and recall. Approx is the number of wordnet-hops allowed.
   * May take long if approx!=0. sym says whether the relation is symmetric. */
  @SuppressWarnings("unchecked")
  public static double[] eval(int approx, boolean sym) {
    Collections.sort(outputPairs);
    double[] result=new double[outputPairs.size()<<1];
    NumberFormatter formi=new NumberFormatter("##.#######");
    int i=0;
    int goodones=0;
    double bestf1=0;
    double bestgoodones=0;
    double bestfound=0;
    int posexamples=0;
    int negexamples=0;
    int bestposexamples=0;
    int bestnegexamples=0;
    System.out.println("----- Output Pairs");
    Announce.progressStart("Evaluating", outputPairs.size());
    for(EvaluatePair p : outputPairs) {
      boolean found=false;
      for(EvaluatePair p2 : idealPairs) {
        if(p2.comment!=null && p.comment!=null && p2.comment.length()>0 && p.comment.length()!=0
              && !p.comment.equalsIgnoreCase(p2.comment)) continue;
        if(approx==0 && p2.equals(p) ||
              approx!=0 && EvaluatePair.strequal(p.word1,p2.word1) && wordNet.hops(p.word2,p2.word2)<approx) {
          goodones++ ;
          found=true;
          p2.hit=true;
          p.hit=true;
          break;
        }
      }
      if(p.example) {
        if(found) posexamples++;
        else negexamples++;
      }
      double precision=(sym?goodones*2:goodones)/(double) (i+1);
      double recall=goodones/(double) idealPairs.size();
      double f1=2*precision*recall/(recall+precision);
      if(f1>bestf1) {
        bestf1=f1;
        bestgoodones=goodones;
        bestfound=i+1;
        bestposexamples=posexamples;
        bestnegexamples=negexamples;
      }
      result[i*2]=precision;
      result[i*2+1]=recall;
      System.out.println(formi.format(precision)+"\t"+formi.format(recall)+"\t"
            +(found ? "OK" : "FAIL")+"\t"+p+" "+f1);
      Announce.progressAt(i++);
    }
    Announce.progressDone();
    System.out.println("--- Ideal Pairs that were not found");
    for(EvaluatePair p2 : idealPairs) {
      if(!p2.hit) System.out.println(p2);
    }
    System.out.println("Best F1:");
    System.out.println("  F1: "+bestf1);
    System.out.println("  Found: "+(sym?bestfound/2:bestfound));
    System.out.println("  Good ones: "+bestgoodones);
    System.out.println("  Positive examples: "+bestposexamples);
    System.out.println("  Negative examples: "+bestnegexamples);
    System.out.println("Total: ");
    System.out.println("  Found: "+(sym?outputPairs.size()/2:outputPairs.size()));
    System.out.println("  Good ones: "+goodones);
    System.out.println("  Positive examples: "+posexamples);
    System.out.println("  Negative examples: "+negexamples);
    System.out.println("Number of ideal pairs: "+idealPairs.size());
    return (result);
  }

  /** Calculates precision and recall according to the ideal metric. Approx
   * is the number of wordnet-hops allowed.
   * May take long if approx!=0. sym says whether the relation is symmetric. */
  @SuppressWarnings("unchecked")
  public static double[] evalIdeal(int approx, boolean sym) {
    // Make idealPairs a multi-valued function
    Map<String,java.util.List<String>> ideal=new HashMap<String,java.util.List<String>>();
    for(EvaluatePair p : idealPairs) {
      java.util.List<String> mapsTo=ideal.get(p.word1);
      if(mapsTo==null) {
        mapsTo=new ArrayList<String>(3);
        mapsTo.add(p.word2);
        ideal.put(p.word1,mapsTo);
        continue;
      }
      if(!mapsTo.contains(p.word2)) mapsTo.add(p.word2);
    }
    // Make outputPairs right-unique
    // Eliminate pairs with an irrelevant first component
    java.util.List<EvaluatePair> output=new ArrayList<EvaluatePair>();
    loop:
    for(EvaluatePair p : outputPairs) {
      // Do we want it?
      if(ideal.get(p.word1)==null) continue;
      // Do we already have it?
      // Replace it if we have a better confidence
      for(EvaluatePair p2 : output) {
        if(EvaluatePair.strequal(p.word1,p2.word1)) {
          if(p2.confidence<p.confidence) {
            p2.confidence=p.confidence;
            p2.word2=p.word2;
          }
          continue loop;
        }
      }
      output.add(p);
    }
    Collections.sort(output);
    double[] result=new double[output.size()<<1];
    NumberFormatter formi=new NumberFormatter("##.#######");
    int i=0;
    int goodones=0;
    double bestf1=0;
    double bestgoodones=0;
    double bestfound=0;
    System.out.println("----- Output Pairs");
    // Run through all outputPairs pairs
    for(EvaluatePair p : output) {
      boolean found=false;
      java.util.List<String> mapsTo=ideal.get(p.word1);
      // Figure out whether what we map to is OK
      for(String s : mapsTo) {
        //int w=0;
        if(approx==0&&EvaluatePair.strequal(s, p.word2)||approx!=0
              &&(wordNet.hops(s, p.word2))<approx) {
          goodones++ ;
          found=true;
          // break keys;
          break;
        }
        //if(approx!=0&&w>=approx) {
        //  D.p(p.word1+": "+p.word2+" instead of "+s+" hops: "+w);
        //}
      }
      double precision=(sym?goodones*2:goodones)/(double) (i+1);
      double recall=goodones/(double) ideal.size();
      double f1=2*precision*recall/(recall+precision);
      if(f1>bestf1) {
        bestf1=f1;
        bestgoodones=goodones;
        bestfound=i+1;
      }
      result[i*2]=precision;
      result[i*2+1]=recall;
      System.out.println(formi.format(precision)+"\t"+formi.format(recall)+"\t"
            +(found ? "OK" : "FAIL")+"\t"+p+" "+f1);
      i++ ;
    }
    System.out.println("----- Full Ideal Set");
    for(String s : ideal.keySet()) {
      System.out.println(s+": "+ideal.get(s));
    }
    System.out.println("Best F1:");
    System.out.println("  F1: "+bestf1);
    System.out.println("  Found: "+(sym?bestfound/2:bestfound));
    System.out.println("  Good ones: "+bestgoodones);
    System.out.println("Total: ");
    System.out.println("  Found: "+(sym?outputPairs.size()/2:outputPairs.size()));
    System.out.println("  Found intersected by Ideal: "+output.size());
    System.out.println("  Good ones: "+goodones);
    System.out.println("Number of ideal pairs: "+ideal.size());
    return (result);
  }

  /** User interface*/
  public static void main(final String[] args) throws Exception {
    if(args.length<2) {
      System.err.println("\n\n   Evaluate [<system>] <option>* <outputFile> <idealFile> ... > <verboseFile>\n");
      System.err.println("Produces a visual representation of the quality of output pairs.");
      System.err.println("<system> - one of "+Arrays.toString(IESystem.values()));
      System.err.println("<option> - ");
      System.err.println("           ideal - use the Snowball Ideal Metric");
      System.err.println("           sym - ignore every second output pair (for symmetry)");
      System.err.println("           <approx> - up to <approx> WordNet hops count as equal.\n");
      System.err.println("<outputFile> - file with the output pairs of the system");
      System.err.println("<idealFile> - files with the ideal pairs");
      System.err.println("              may contain wildcards (quote the filename)");
      System.err.println("<verboseFile> - file where the results of the evaluation are written to");
      System.exit(63);
    }
    boolean sym=false;
    IESystem system=IESystem.LEILA;
    int approx=0;
    boolean ideal=false;
    String output=null;
    Announce.doing("Mounting");
    for(String s : args) {
      if(s.equalsIgnoreCase("ideal")) {
        ideal=true;
        continue;
      }
      if(s.equalsIgnoreCase("sym")) {
        sym=true;
        continue;
      }
      if(s.matches("\\d+")) {
        approx=Integer.parseInt(s);
        Announce.doneDoing("Loading WordNet");
        wordNet=new InstanceOf();
        continue;
      }
      try {
        system=IESystem.valueOf(s.toUpperCase());
        continue;
      } catch(Exception e) {}
      if(output==null) {
        output=s;
        continue;
      }
      for(File f : new DeepFileSet(s)) {
        Announce.doneDoing("Reading ideal pairs "+f);
        addIdeal(f);
      }
    }
    Announce.done();
    System.out.println("\n\n"+system+" Evaluation");
    System.out.println("  Time: "+NumberFormatter.ISOtime());
    System.out.println("  Options:");
    System.out.println("    Ideal: "+ideal);
    System.out.println("    Symmetric: "+sym);
    System.out.println("    Approximation: "+approx);
    System.out.println("  OutputFile: "+output);
    System.out.println("  IdealFiles: "+files+"\n\n");
    addOutput(new File(output),system);
    double[] a;
    if(ideal) a=evalIdeal(approx,sym);
    else a=eval(approx,sym);
    Announce.doing("Opening window");
    JFrame f=new JFrame("Evaluate");
    f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    JApplet applet=new Evaluate(a);
    f.getContentPane().add("Center", applet);
    applet.init();
    f.pack();
    f.setVisible(true);
    Announce.done();
  }
}