package leila.evaluate;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;

import javatools.parsers.Char;
import javatools.parsers.Name;
import javatools.parsers.NumberFormatter;
import javatools.parsers.PlingStemmer;

import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JEditorPane;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.event.CaretEvent;
import javax.swing.event.CaretListener;
import javax.swing.text.DefaultHighlighter;
import javax.swing.text.Highlighter;


/**
This class is part of LEILA (http://mpii.de/yago-naga/leila).
It is licensed under the Creative Commons Attribution License 
(http://creativecommons.org/licenses/by/3.0) 
by the author Fabian M. Suchanek (http://suchanek.name).

<P>

HandTag provides a visual interface for hand-tagging the elements of a binary relation
in an HTML-document.
 */
public class HandTag extends JPanel {

    /** Holds the current first word */
    protected Word firstWord=new Word();
    /** Holds the current second word */
    protected Word secondWord=new Word();
    /** Points to the message label */
    protected JLabel info=new JLabel(" ");
    /** Editor pane */
    protected JEditorPane editorPane=new JEditorPane();
    /** Highlighter for marked words */
    protected Highlighter hili=new DefaultHighlighter();
    /** Undo button */
    protected JButton undo=new JButton("Undo");
    /** Upper panel */
    protected JPanel northPanel=new JPanel();
    /** Writer where tagged pairs are written to */
    protected static BufferedWriter out;

    /** Represents one marked word*/
    protected class Word {
      /** Used to highlight the word*/
      protected Highlighter.HighlightPainter hilipa=new DefaultHighlighter.DefaultHighlightPainter(Color.CYAN);
      /** Start position in text */
      public int start;
      /** End position in text */
      public int len;
      /** Identifies the highlighter*/
      protected Object h;

      /** Tells whether this word is empty*/
      public boolean empty() {
        return(start==-1);
      }

      /** Returns the word as String*/
      public String getWord() {
        if(empty()) return(null);
        try {
         String s=editorPane.getText(start,len).replace('\n',' ').trim();
         if(!Name.isName(s)) s=PlingStemmer.stem(s);
         return(s);
        } catch(Exception e) {
         return(null);
        }
      }

      /** Resets this word*/
      public void clear(){
        if(empty()) return;
        start=-1;
        if(hili!=null) hili.removeHighlight(h);
      }

      /** Constructor */
      public Word() {
        clear();
      }

      /** Set this word by giving start and end position in text */
      public void set(int dot, int mark) {
        clear();
        start=mark;
        len=dot-mark;
        try{
         while(editorPane.getText(start,len).length()==0 ||
               Character.isLetterOrDigit(editorPane.getText(start-1,len).charAt(0))) {
           start--;
           len++;
         }
         while(Character.isLetterOrDigit(Char.last(editorPane.getText(start,len+1)))) {
           len++;
         }
         h=hili.addHighlight(start,start+len,hilipa);
        } catch(Exception e) {
          start=-1;
        }
      }
    }

    /** New Handtagger for a file */
    public HandTag(String f) {
        setLayout(new BorderLayout());
        //Create an editor pane.
        try {         f=new File(f).getCanonicalPath();
         editorPane.setPage("file:///"+f.replace('\\','/')); } catch(Exception e){System.exit(255);};
        editorPane.setEditable(false);
        editorPane.setHighlighter(hili);
        editorPane.setSelectionColor(Color.CYAN);
        editorPane.addCaretListener(new CaretListener() {
          public void caretUpdate(CaretEvent e) {
            if(secondWord.empty() && !firstWord.empty()) {
             secondWord.set(e.getDot(), e.getMark());
             updateInfo();
            } else {
             push();
             firstWord.set(e.getDot(), e.getMark());
             secondWord.clear();
             if(firstWord.getWord().equalsIgnoreCase("he") || firstWord.getWord().equalsIgnoreCase("she")) {
               info.setText("Please click the name of the person instead of the pronoun!");
               firstWord.clear();
             } else if(firstWord.getWord().length()==0 ||
                   !Character.isLetterOrDigit(firstWord.getWord().charAt(0))) {
               // Do nothing
               firstWord.clear();
               info.setText("You clicked on an invalid position in the document");
             } else {
               updateInfo();
             }
            }
          }
        });
        JScrollPane editorScrollPane = new JScrollPane(editorPane);
        editorScrollPane.setVerticalScrollBarPolicy(
                        JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
        editorScrollPane.setPreferredSize(new Dimension(1000, 800));
        add(editorScrollPane,BorderLayout.CENTER);
        undo.addActionListener(new ActionListener() {
           public void actionPerformed(ActionEvent e) {
             if(secondWord.empty()) {
               firstWord.clear();
             } else {
               secondWord.clear();
             }
             updateInfo();
           }
        });
        northPanel.setLayout(new BorderLayout());
        northPanel.add(info,BorderLayout.WEST);
        northPanel.add(undo,BorderLayout.EAST);
        add(northPanel,BorderLayout.NORTH);
    }

    /** Updates the info bar*/
    public void updateInfo() {
     if(firstWord.empty()) info.setText(" ");
     else if(secondWord.empty())  info.setText(firstWord.getWord());
     else info.setText(firstWord.getWord()+" / "+secondWord.getWord());
    }

    /** Writes the current pair to the output writer*/
    public void push() {
      if(!secondWord.empty()) {
        try{
         out.write(firstWord.getWord()+" / "+secondWord.getWord()+"\n");
        }catch(Exception e){};
      }
    }

    /** Create the GUI and show it */
    protected static void createAndShowGUI(String f) {

        //Create and set up the window.
        JFrame frame = new JFrame("HandTag");
        //Create and set up the content pane.
        final JComponent newContentPane = new HandTag(f);
        frame.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
        frame.addWindowListener(new WindowAdapter() {
          public void windowClosing(WindowEvent e) {
            ((HandTag)newContentPane).push();
            try{out.close();}catch(Exception x){};
            System.exit(0);
          }
        });
        newContentPane.setOpaque(true); //content panes must be opaque
        frame.setContentPane(newContentPane);

        //Display the window.
        frame.pack();
        frame.setVisible(true);
    }

    /** Mounts the main program */
    public static void main(final String[] args) throws Exception {
        if(args.length<2) {
          System.err.println("\n\n   HandTag <corpusFile> <idealFile> \n");
          System.err.println("Lets you select relation pairs manually from a corpus file.");
          System.err.println("<corpusFile> - corpus file (html or txt)");
          System.err.println("<idealFile> - file where the ideal pairs will be written to\n");
          System.err.println("When the program runs, just mark the first component and then the second");
          System.err.println("component of an ideal pair. As soon as you continue with a new pair");
          System.err.println("(or you quit the program), the previous pair is saved.");
          System.exit(63);

        }
        out=new BufferedWriter(new FileWriter(args[1]));
        out.write("% Manually selected ideal pairs\n");
        out.write("%   File: "+args[0]+"\n");
        out.write("%   Time: "+NumberFormatter.ISOweekTime()+"\n");
        //Schedule a job for the event-dispatching thread:
        //creating and showing this application's GUI.
        javax.swing.SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                createAndShowGUI(args[0]);
            }
        });
    }
}