package leila.parsing;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.util.Calendar;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javatools.parsers.NumberFormatter;
/**
This class is part of LEILA (http://mpii.de/yago-naga/leila).
It is licensed under the Creative Commons Attribution License 
(http://creativecommons.org/licenses/by/3.0) 
by the author Fabian M. Suchanek (http://suchanek.name).

<P>

The DateParser normalizes date expressions in english natural language text
to expressions of the form <BR>
<CENTER>&#064;<i>year_month_day</i>&#064;</CENTER><BR>
where <i>year</i> is either positive negative.
The DateParser understands expressions like "4th century BC" or "3rd of November 2004".
Dates may be underspecified: The character '#' stands for "at least one digit".
A match-function is provided that checks if two potentially underspecified dates
or year numbers match. Every positive normalized integer number is a date (namely a year).<BR>
Example:
   <PRE>
         DateParser.normalize("It was November 23rd to 24th 1998.")
         --> "It was &#064;1998_11_23&#064; to &#064;1998_11_24&#064;."
         NumberFormatter.ISOtime(DateParser.getCalendar("November 24th 1998"))
         --> 1998-12-24 T 00:00:00.00
   </PRE>
*/

public class DateParser {

  /** Creates a date-string of the form &#064;<i>year_month_day</i>&#064;
    * from a day, month and year */
  public static final String newDate(String d,String m,String y) {
    return("@"+y+"_"+m+"_"+d+'@');
  }

  /** Creates a date-string from a day, month and year as ints */
  public static final String newDate(int d,int m,int y) {
    return(newDate(""+d,""+m,""+y));
  }

  /** The general default date */
  public static final String ANYDATE=newDate(0,0,0);

  /** A Date as a capturing RegEx */
  public static final String DATE=newDate("([0-9#]+)","([0-9#]+)","(-?[0-9#]+)");
  public static final Pattern DATEPATTERN=Pattern.compile(DATE);

  /** Just a pair of a Pattern and a replacement string */
  private static class FindReplace {
    public Pattern pattern;
    public String replacement;
    public FindReplace(String f,String r) {
      pattern=Pattern.compile(f);
      replacement=r;
    }
  }
  /** Contains the month short names */
  private static final String MONTHS="JanFebMarAprMayJunJulAugSepOctNovDec";

  /** A blank as a RegEx */
  private static final String B="[\\W_&&[^-]]*";
  /** A forced blank as a RegEx */
  private static final String FB="[\\W_&&[^-]]+";
  /** A hyphen as a RegEx with blanks*/
  private static final String H=B+"(?:-+|to|until)"+B;
  /** BC as a RegEx with blank*/
  private static final String BC=B+"(?:BC|B\\.C\\.|BCE|AC|A\\.C\\.)";
  /** AD as a RegEx with blank*/
  private static final String AD="AD|A\\.D\\."+B;
  /** CE as a RegEx with blank*/
  private static final String CE=B+"CE|C\\.E\\.";
  /** ##th as a capturing RegEx with blank*/
  private static final String NTH="(\\d{1,2})[a-z]{2}"+B;
  /** ##[th] as a capturing RegEx*/
  private static final String N="(\\d{1,2})[a-z]{0,2}";
  /** #### as a capturing RegEx */
  private static final String Y4="(\\d{4})";
  /** ###...# as a capturing RegEx */
  private static final String Y="(\\d{1,10})";
  /** MONTH## as a capturing RegEx */
  private static final String M="MONTH(\\d\\d)";
  /** A "the" as a RegEx with blank*/
  private static final String THE="(?:the)?"+B;
  /** century as a RegEx */
  private static final String CENTURY="[cC]entur(?:y|(?:ies))";
  /** millenium as a RegEx */
  private static final String MILENNIUM="[mM]ill?enn?ium";

  /** Holds the date patterns */
  // The internal order of the patterns is essential!
  //    We always process periods of time first in order to capture the part of it
  //    that lacks explicit date identification
  //    Furthermore, we first process "BC" because we might loose it else
  private static final FindReplace[] patterns=new FindReplace[]{

    //  --------- Process ISO8601 ------------------
    new FindReplace(Y+"-(\\d{1,2})-(\\d{1,2})",newDate("$3","$2","$1")),

    //  --------- Process BC, CE and AD ------------
    // 2267 - 2213 BC
    new FindReplace(Y+H+Y+BC,newDate("#","#","-$1")+" to "+newDate("#","#","-$2")),
    // 2267 - 2213 CE
    new FindReplace(Y+H+Y+CE,newDate("#","#","$1")+" to "+newDate("#","#","$2")),
    // 1000 BC - AD 120
    new FindReplace(Y+BC+H+AD+Y,newDate("#","#","-$1")+" to "+newDate("#","#","$2")),
    // 1000 BC - 120 CE
    new FindReplace(Y+BC+H+Y+CE,newDate("#","#","-$1")+" to "+newDate("#","#","$2")),
    // AD 46 - 120
    new FindReplace(AD+Y+H+Y,newDate("#","#","$1")+" to "+newDate("#","#","$2")),
    // 1000 BC
    new FindReplace(Y+BC,newDate("#","#","-$1")),
    // 1000 CE
    new FindReplace(Y+CE,newDate("#","#","$1")),
    // AD 1000
    new FindReplace(AD+Y,newDate("#","#","$1")),

    //  --------- Process complete dates ----------
    // 23rd - 24th of November 1998
    new FindReplace(FB+N+H+N+B+"(?:of)?"+B+M+B+Y," "+newDate("$1","$3","$4")+" to "+newDate("$2","$3","$4")),
    // November 23rd - 24th 1998
    new FindReplace(M+B+N+H+N+FB+Y,newDate("$2","$1","$4")+" to "+newDate("$3","$1","$4")),
    // November 23rd - March 24th 1998
    new FindReplace(M+B+N+H+M+B+N+FB+Y,newDate("$2","$1","$5")+" to "+newDate("$4","$3","$5")),
    // November 23[rd] 1998
    new FindReplace(M+B+N+FB+Y,newDate("$2","$1","$3")),
    // November 23, 1998
    new FindReplace(M+B+N+B+","+B+Y,newDate("$2","$1","$3")),
    // 23[rd] November 1998
    new FindReplace(N+B+M+B+Y,newDate("$1","$2","$3")),
    // 23[rd] of November 1998
    new FindReplace(N+B+"of"+B+M+B+Y,newDate("$1","$2","$3")),
    // 12-Sep-1970
    new FindReplace(N+H+M+H+Y,newDate("$1","$2","$3")),

    //  --------- Process dates with months ----------
    // June - April 1980
    new FindReplace(M+H+M+B+Y4,newDate("#","$1","$3")+" to "+newDate("#","$2","$3")),
    // April 1980
    new FindReplace(M+B+Y4,newDate("#","$1","$2")),

    //  --------- Process days of the year ----------
    // June 14th - July 17th
    new FindReplace(M+B+N+H+M+B+N,newDate("$2","$1","#")+" to "+newDate("$4","$3","#")),
    // June 14th - 17th
    new FindReplace(M+B+N+H+N,newDate("$2","$1","#")+" to "+newDate("$3","$1","#")),
    // June 14th
    new FindReplace(M+B+N,newDate("$2","$1","#")),

    // ----------- Process centuries (add subtraction marker) -----------
    // ##th - ##th millennium BC
    new FindReplace(THE+NTH+H+NTH+B+MILENNIUM+BC,' '+newDate("#","#","-&DEC$1###")+" to "+newDate("#","#","-&DEC$2###")),
    // ##th - ##th century BC
    new FindReplace(THE+NTH+H+NTH+B+CENTURY+BC,' '+newDate("#","#","-&DEC$1##")+" to "+newDate("#","#","-&DEC$2##")),
    // ##th - ##th century
    new FindReplace(THE+NTH+H+NTH+B+CENTURY,' '+newDate("#","#","&DEC$1##")+" to "+newDate("#","#","&DEC$2##")),
    // ##th millennium BC
    new FindReplace(THE+NTH+B+MILENNIUM+BC,' '+newDate("#","#","-&DEC$1###")),
    // ##th century BC
    new FindReplace(THE+NTH+B+CENTURY+BC,' '+newDate("#","#","-&DEC$1##")),
    // ##th millennium
    new FindReplace(THE+NTH+B+MILENNIUM,' '+newDate("#","#","&DEC$1###")),
    // ##th century
    new FindReplace(THE+NTH+B+CENTURY,' '+newDate("#","#","&DEC$1##")),

    // ------------ Process special constructions -----------
    // 1850s
    new FindReplace("(\\d{3})0'?s",newDate("#","#","$1#")),
    // about date
    new FindReplace("("+B+")"+"(?:about|circa|c\\.a\\.|ca\\.|c|c\\.)"+B+DATE,"$1"+newDate("$2","$3","$4"))
  };
  /** Holds the pattern seeking for month names */
  private static final Pattern monthPattern=Pattern.compile(
    "(Jan|January|Feb|February|Febr|Mar|March|Apr|April|May|Jun|June|Jul|July|Aug|August|Sep|September|Sept"+
    "|Oct|October|Nov|November|Dec|December)[\\W_]");
    // May terminate with '_' because Wikipedia highlights days of month by <A>...</A>
    // (like <A>May 3</A>) and HTML2LGI joins A-Tags by '_'.
  /** Holds the pattern that determines whether a string contains a year expression */
  private static final Pattern yearPattern=Pattern.compile(
        "\\d{2,4}|\\d{4}s|B\\.?C\\.?|A\\.?C\\.?|A\\.?D\\.?|"+CENTURY+"|\\d{2,4}s");
  /** Normalizes all dates in a StringBuilder, may mess up the original StringBuilder */
  public static StringBuilder normalize(StringBuilder s) {
    Matcher m=monthPattern.matcher(s);
    // Check out if a month appears
    if(m.find()) {
      // While there are months in the String, normalize them to MONTH##
      do{
        int monthNum=MONTHS.indexOf(m.group().substring(0,3))/3+1;
        s.replace(m.start(),m.end()-1,"MONTH"+(monthNum/10)+(monthNum%10));
        m=monthPattern.matcher(s);
      }while(m.find());
    } else {
      // No month found, if no year is in it, drop it
      if(!yearPattern.matcher(s).find()) return(s);
    }
    // Test patterns
    for(int i=0;i<patterns.length;i++) {
      m=patterns[i].pattern.matcher(s);
      if(m.find()) {
         s=new StringBuilder(m.replaceAll(patterns[i].replacement));
      }
    }
    // Test subtractions
    int i=-1;
    while((i=s.indexOf("&DEC",i))!=-1) {
      int j=s.indexOf("#",i);
      s.replace(i,j,""+(Integer.parseInt(s.substring(i+4,j))-1));
    }
    // Cut out leading zeroes
    while((i=s.indexOf("_0",i))!=-1) {
      s.delete(i+1,i+2);
    }
    return(s);
  }

  /** Normalizes all dates in a String
    * Convenience method for the more efficient StringBuilder-method */
  public static String normalize(String s) {
    return(normalize(new StringBuilder(s)).toString());
  }

  /** Tells whether this string is a normalized date (or year number) */
  public static boolean isDate(String s) {
    return(s.matches(DATE) || NumberParser.isInt(s));
  }

  /** Tells whether two date-strings match
    * Also matches normalized positive year numbers with other positive year numbers or dates */
  public static boolean match(String d1, String d2) {
    return(
          d1.equals(ANYDATE) && isDate(d2)
       || d2.equals(ANYDATE) && isDate(d1)
       || d1.endsWith("_#_#@") && NumberParser.match(NumberParser.newNumber(d1.substring(1,d1.length()-5)),d2)
       || d2.endsWith("_#_#@") && NumberParser.match(NumberParser.newNumber(d2.substring(1,d2.length()-5)),d1)
       || NumberParser.isInt(d1) && NumberParser.match(d1,d2)
       || isDate(d1) && d1.equals(d2));
  }

  /** Extracts the year as a String */
  public static String getYear(String d) {
    Matcher m=DATEPATTERN.matcher(d);
    if(m.find()) return(m.group(1));
    return("");
  }
  
  /** Extracts the month as a String */
  public static String getMonth(String d) {
    Matcher m=DATEPATTERN.matcher(d);
    if(m.find()) return(m.group(2));
    return("");
  }

  /** Extracts the day as a String */
  public static String getDay(String d) {
    Matcher m=DATEPATTERN.matcher(d);
    if(m.find()) return(m.group(3));
    return("");
  }
  
  /** Converts a String to an integer or 0 in case of failure */
  public static int toInt(String s) {
    int i=0;
    try {
      i=Integer.parseInt(s);
    }
    catch(NumberFormatException e) {
    }
    return(i);
  }
  
  /** Converts a String to a Calendar */
  public static Calendar getCalendar(String s) {
    s=normalize(s).replace('#', '0');
    Calendar c=Calendar.getInstance();
    c.set(toInt(getYear(s)), toInt(getMonth(s)), toInt(getDay(s)),0,0,0);
    return(c);
  }
  
  /** Test routine */
  public static void main(String[] argv) throws Exception {
    System.out.println(NumberFormatter.ISOtime(DateParser.getCalendar("November 24th 1998")));         
    System.out.println("Enter a string containing a date expression and hit ENTER. Press CTRL+C to abort");
    while(true) {
      String in=new BufferedReader(new InputStreamReader(System.in)).readLine();
      System.out.println(normalize(in));
      System.out.println(getCalendar(in));
    }
  }

}
