package leila.parsing;
import java.util.Map;
import java.util.TreeMap;

import javatools.datatypes.FinalSet;
import javatools.parsers.Name;
import leila.patterns.Link;
import leila.patterns.LinkageWord;

/**
This class is part of LEILA (http://mpii.de/yago-naga/leila).
It is licensed under the Creative Commons Attribution License 
(http://creativecommons.org/licenses/by/3.0) 
by the author Fabian M. Suchanek (http://suchanek.name).

<P>

Document.java maintains the contextual information of a sentence, like
pronoun references and full names of people.
 */
public class Document {
  /** Number of sentences during which a pronoun reference remains valid.
   *  A value of 2 means: current and following sentence   */
  protected static final int SPAN=2;
  /** Contains words that refer to companies */
  protected static final FinalSet<String> companies=new FinalSet<String>(new String[] {
     "company",
     "maker"
  });
  /** Counts the sentences */
  protected int current=0;
  /** Contains the mapping familiyname/companyname->fullname */
  protected Map<String,LinkageWord> names=new TreeMap<String,LinkageWord>();
  /** Represents a reference target, memorizes the time it has been created */
  protected class AgedLinkageWord {
    /** Stores the sentence number at this AgedLinkageWord's creation */
    public int birthday;
    /** Stores the linkage word */
    public LinkageWord linkageWord;
    /** Constructor */
    public AgedLinkageWord(LinkageWord l) {
      birthday=current;
      linkageWord=l;
    }
    /** Sets the birthday to now */
    public void refresh() {
      birthday=current;
    }
    /** Tells whether this AgedLinkageWord's creation is less than SPAN ago */
    public boolean isAlive() {
      return(current-birthday<SPAN);
    }
    /** Returns a simple String */
    public String toString() {
      return(linkageWord.toString());
    }
  }
  /** Maps references to their targets */
  protected Map<String,AgedLinkageWord> references=new TreeMap<String,AgedLinkageWord>();

  /** Notification for beginning of a new sentence */
  public void newSentence() {
    current++;
  }

  /** Tells whether the LinkageWord is a resolvable reference */
  public static boolean isResolvableReference(LinkageWord w) {
    // Pronouns are resolvable
    if(w.is(LinkageWord.Type.PRONOUN)) return(true);
    // Check if this is a company substitute
    if(!companies.contains(w.word.substring(w.word.lastIndexOf('_')+1))) return(false);
    //  If it is a company substitute, reject it if it is isPlural
    if(w.is(LinkageWord.Type.PLURAL)) return(false);
    // If it is a company substitute, reject it if it is "a company"
    for(Link l : w.contexts) {
      if(l.destination.word.equals("a")) return(false);
    }
    // Else accept it
    return(true);
  }

  /** Resolves a reference to its target */
  public void resolveReference(LinkageWord w) {
    if(!isResolvableReference(w)) return;
    AgedLinkageWord result=references.get(w.word);
    if(result==null) result=references.get(w.word.substring(w.word.lastIndexOf('_')+1));
    if(result==null || !result.isAlive()) return;
    w.adoptWord(result.linkageWord);
  }

  /** Adds a LinkageWord to the corresponding pronoun(s) */
  public void addTarget(LinkageWord l) {
    if(l.is(LinkageWord.Type.PLURAL)) {
      references.put("they", new AgedLinkageWord(l));
      return;
    }
    if(l.is(LinkageWord.Type.COMPANY) || l.is(LinkageWord.Type.COULDBEPERSON)) {
      AgedLinkageWord a=new AgedLinkageWord(l);
      for(String s : companies) references.put(s,a);
    }
    if(l.is(LinkageWord.Type.NAME)) {
      // Do not store a referee for "it", because "it" appears in many phrases
      AgedLinkageWord a=new AgedLinkageWord(l);
      references.put("he", a);
      references.put("she", a);
      references.put("him", a);
    }
  }

  /** Refreshes a target */
  public void refreshTargetFor(String s) {
    AgedLinkageWord a=references.get(s);
    if(a!=null) a.refresh();
  }

  /** Adds a completion mapping familyName/companyName -> fullName */
  public void addName(LinkageWord l) {
    if(l.is(LinkageWord.Type.PERSON)) {
      String family=new Name.PersonName(l.word).familyName();
      String normalized=new Name.PersonName(l.word).normalize();
      if(family!=null && family.length()<normalized.length()) {
        l.word=normalized;
        names.put(family,l);
      }
    } else if(l.is(LinkageWord.Type.COMPANY) && l.word.lastIndexOf('_')!=-1) {
      String cname=l.word.substring(0,l.word.lastIndexOf('_'));
      if(cname.length()>1) names.put(cname,l);
    }
  }
  /** Completes a family name/company name by the full name if the name is known. */
  public void completeName(LinkageWord w) {
    LinkageWord result=names.get(w.word);
    if(result!=null) {
      w.adoptWord(result);
      return;
    }
    String family=new Name.PersonName(w.word).familyName();
    if(family==null || family.length()!=w.word.length()) return;
    result=names.get(family);
    if(result==null) return;
    w.adoptWord(result);
  }
}