package leila.parsing;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.Writer;
import java.util.regex.Pattern;

import javatools.administrative.Announce;
import javatools.datatypes.FinalSet;
import javatools.filehandlers.DeepFileSet;
import javatools.parsers.Char;
import leila.patterns.LinearLinkageFile;
import leila.patterns.LinkageFile;


/**
This class is part of LEILA (http://mpii.de/yago-naga/leila).
It is licensed under the Creative Commons Attribution License 
(http://creativecommons.org/licenses/by/3.0) 
by the author Fabian M. Suchanek (http://suchanek.name).

<P>

HTML2LGI takes HTML-files and produces LGI-files (Link Grammar Input files)
and LL-files (Linear Linkage files, non-grammatical part).
*/

public class HTML2LGI {
  /** Maximal length of a word, must be <60,
   * which is the maximum that the link grammar parser can swallow */
  public static final int MAXJOIN=40;
  /** Readers and Writers */
  protected static HTMLReader in;
  protected static Writer lgiout;
  protected static Writer llout;
  /** Minimal sentence length (in chars) */
  protected static final int MINLGILENGTH=25;
  /** Minimal linear linkage length (in chars) */
  protected static final int MINLLLENGTH=5;
  /** Maximal linear linkage length (in chars) */
  protected static final int MAXLLLENGTH=50;
  /** Abbreviations */
  public static FinalSet<String> abbreviations=new FinalSet<String>(
        "Apr",
        "Aug",
        "Corp",
        "Dec",
        "Dr",
        "Drs",
        "Feb",
        "Febr",
        "Inc",
        "Jan",
        "Jr",
        "Jul",
        "Jun",
        "Ltd",
        "Mar",
        "Mr",
        "Mrs",
        "Ms",
        "No",
        "Nov",
        "Oct",
        "Sep",
        "Sept",
        "Sr",
        "jr",
        "ltd",
        "sr"
  );
  /** Characters to be ignored*/
  public static final String ignoreChars="{}[]\"";
  /** HTML-tags to be ignored */
  public static FinalSet<String> ignores=new FinalSet<String>(
        "/B",
        "/FONT",
        "/STRONG",
        "/SUP",
        "/U",
        "B",
        "FONT",
        "STRONG",
        "U"
  );
  /** HTML-tags to be skipped */
  public static FinalSet<String> skips=new FinalSet<String>(
        "ADDRESS",
        "BDO",
        "CITE",
        "CODE",
        "KBD",
        "LISTING",
        "PLAINTEXT",
        "PRE",
        "Q",
        "SAMP",
        "TT",
        "VAR",
        "XMP"
  );
  /** Characters that count as sentence delimiters */
  public static final String stopChars="!.;|?:";
  /** HTML-tags that count as sentence delimiters */
  public static FinalSet<String> stops=new FinalSet<String>(
        "/DD",
        "/H1",
        "/H2",
        "/H3",
        "/H4",
        "/H5",
        "/H6",
        "/P",
        "/TR",
        "BR",
        "BR/",
        "DD",   // This is used for definitions, but not in Wikipedia
        "P",
        "TR"
  );
  /** HTML-tags that emphasize a string*/
  public static final FinalSet<String> emphStart=new FinalSet<String>(
      "A",
      "EM",
      "I"
  );
  /** HTML-tags that finish the emphasis of a string*/
  public static final FinalSet<String> emphEnd=new FinalSet<String>(
    "/A",
    "/EM",
    "/I"
  );

  /** Flushes data to LGIfile */
  protected static void flushLGI(String s) throws Exception {
      // Use one more space to separate dot from preceding date/number
      lgiout.write(NumberParser.normalize(DateParser.normalize(s)));
      lgiout.write(" .\n\n\n");
  }

  /** Flushes data to LLfile*/
  protected static void flushLL(String s) throws Exception {
    llout.write(NumberParser.normalize(DateParser.normalize(s)));
    llout.write("\n");
  }

  /** Flushes data, resets s */
  protected static void flush(String headline, StringBuilder s) throws Exception {
    if(s.length()>MINLGILENGTH) flushLGI(s.toString());
    if(s.length()>MINLLLENGTH && s.length()<MAXLLLENGTH
          && headline.length()>3) {
      flushLL(headline+s);
    }
    s.setLength(0);
  }

  /** Returns a sequence of unproblematic characters and its delimiter */
  protected static StringBuilder readSentence(String[] delim) throws Exception {
    String c=null;
    StringBuilder result=new StringBuilder();
    boolean join=false;
    int joinstart=0;
    boolean cread=false;
    while(true) {
      if(!cread) c=in.readString();
      cread=false;
      // Process EOF
      if(c==null) {
        delim[0]=c=".";
        if(result.length()==0) return (null);
        else break;
      }
      // Process whitespaces
      if(c.charAt(0)==' ') {
        // Skip leading blanks
        if(result.length()==0) continue;
        // Erase duplicate blanks
        if(Char.last(result)==' ') continue;
        result.append(' ');
        continue;
      }
      // Process Tags
      if(c.startsWith("TAG_")) {
        if(ignores.contains(c.substring(4))) continue;
        if(emphStart.contains(c.substring(4))) {
          join=true;
          joinstart=result.length();
          continue;
        }
        if(emphEnd.contains(c.substring(4))) {
          if(join && result.length()-joinstart<MAXJOIN) {
            for(int i=joinstart;i<result.length()-2;i++) {
              if(result.charAt(i)==' ') result.replace(i, i+1, "_");
            }
          }
          join=false;
          continue;
        }
        if(skips.contains(c.substring(4))) {
          String seek="TAG_/"+c.substring(4);
          while(c!=null&&!c.equals(seek)) {
            c=in.readString();
          }
          c=".";
          break;
        }
        if(c.equals("TAG_SUP")) {
          result.append('^');
          continue;
        }
        if(c.equals("TAG_SUB")) {
          result.append('[');
          continue;
        }
        if(c.equals("TAG_/SUB")) {
          result.append(']');
          continue;
        }
        break;
      }
      // Process URLs
      if(c.charAt(0)==':'&&result.toString().endsWith("http")) {
        while(!c.equals(" ")&&!c.startsWith("TAG")&&!c.equals("\n")&&c!=null)
          c=in.readString();
        result.replace(result.length()-4, result.length(), "URL");
        cread=true;
        continue;
      }
      // Process dots
      if(c.charAt(0)=='.') {
        // Process abbreviations like "c."
        if((result.length()>=2 && result.charAt(result.length()-2)==' ' || result.length()==1) &&
              Character.isLetter(Char.last(result))) continue;
        // Process abbreviations like "B.C."
        if(Character.isUpperCase(Char.last(result))) continue;
        // Process known abbreviations
        String last=result.substring(result.lastIndexOf(" ")+1);
        if(abbreviations.contains(last)) continue;
        // Look ahead
        c=in.readString();
        cread=true;
        // Process dots that are followed by a blank
        if(c==null||c.equals(" ")||c.equals("\"")) {
          c=".";
          break;
        }
        // Process number dots
        if(Character.isDigit(c.charAt(0))) {
          result.append('.');
          continue;
        }
        // All other dots can be ignored
        // e.g. "Google.com", "Inc.,", "Inc.</strong>"
        continue;
      }
      // Ignore double quotes
      if(c.equals("'")) {
        if(Char.last(result)=='\'') result.setLength(result.length()-1);
        else result.append("'");
        continue;
      }
      // Process unknowns
      if(c.equals(Char.UNKNOWN)) continue;
      // Process externalized references like "[1]"
      if(c.charAt(0)=='[') {
        while(c!=null&&c.charAt(0)!=']'&&c.charAt(0)!=' ')
          c=in.readString();
        result.append(' ');
        continue;
      }
      // Process Sentence delimiters
      if(stopChars.indexOf(c.charAt(0))!=-1 || c.equals("--")) break;
      if(c.equals("-") && Char.last(result)=='-') {
        result.setLength(result.length()-1);
        break;
      }
      // Igore ignorables
      if(ignoreChars.indexOf(c.charAt(0))!=-1) continue;
      // Process commas in join mode
      if(join&&c.charAt(0)==',') continue;
      // Process other delimiters
      if("()".indexOf(c.charAt(0))!=-1) break;
      result.append(c);
    }
    delim[0]=c;
    // Kill objects that are too long
    if(Pattern.matches(".*[^ ]{"+MAXJOIN+",}.*", result)) result.setLength(0);
    // Trim trailing blank
    if(result.length()>1
          && Char.last(result)==' ') result.setLength(result.length()-1);
    return (result);
  }

  /** Parses a text under a given headline */
  protected static void parseText(String headline) throws Exception {
      String[] delim=new String[1];
      StringBuilder line=new StringBuilder();
      StringBuilder chunk;
      String lastReasonable="";   
      while((chunk=readSentence(delim))!=null) {
        line.append(chunk);
        // Headline
        // * register as headline
        // * generate a pseudo-sentence to identify the headline in the LGI-file
        if(delim[0].matches("TAG_/H\\d")
              && line.length()>3) {
          flushLGI(line+" hobnobs");
          headline=line+" > ";
          if(line.length()>MINLLLENGTH) lastReasonable=line.toString();
          flush("",line);
          continue;
        }
        // Closing parenthesis
        if(delim[0].equals(")")) {
          flush(headline,line);
          if(Char.last(headline)=='(') return;
          // Else we have a mismatched or too distant parenthesis
          // (that's life)
          continue;
        }
        // Opening parenthesis
        if(delim[0].equals("(")) {
          if(line.length()>0) parseText(line.toString()+" (");
          else parseText(lastReasonable+" (");
          line.append(' ');
          continue;
        }
        // List start
        if(delim[0].equals("TAG_OL") || delim[0].equals("TAG_UL")) {
          flush(headline,line);
          parseText(lastReasonable+" # ");
          continue;
        }
        // List end
        if(delim[0].equals("TAG_/OL") || delim[0].equals("TAG_/UL")) {
            flush(headline,line);
            return;
        }
        // Flush first sentence in any case to LL, because it may refer to
        // the headline and be incomplete
        if(headline.endsWith(" > ") && headline.equals(lastReasonable+" > ") &&
              line.length()>MINLLLENGTH) {
          flushLL(headline+line);
          headline="";
        }
        // Save reasonable line as future headline
        // Save as string since <line> will be modified
        if(line.length()>MINLLLENGTH) lastReasonable=line.toString();
        flush(headline,line);
      }
   }

  /** Parses an HTML-file */
  protected static void parseFile(File f) throws Exception {
    in=new HTMLReader(new FileReader(f));
    lgiout=new FileWriter(LinkageFile.lgiFile(f));
    llout=new FileWriter(LinearLinkageFile.llFile(f));
    lgiout.write("!verbosity=0\n");
    lgiout.write("!width=1000\n");
    lgiout.write("!union=1\n");
    // Skip everything to the body
    String z="";
    while(z!=null && !z.equals("TAG_BODY") && !z.equals("TAG_TEXT")) z=in.readString();
    parseText("");
    in.close();
    lgiout.write("quit\n");
    lgiout.close();
    llout.close();
  }

  /** Translates HTML-files to a LGI-file (call by Java)*/
  public static void main(File f) throws Exception {
      Announce.doing("Producing an LGI-file for "+f.getName());
      parseFile(f);
      Announce.done();
  }

  /** Translates HTML-files to a LGI-file (call by User)*/
  public static void main(String[] argv) throws Exception {
    if(argv.length<1) {
      System.err.println("\n\n   HTML2LGI <HTMLfile> \n");
      System.err.println("Translates the HTML-file to an LGI-file (Link Grammar Input)");
      System.err.println("and an LL-file (Linear Linkage, non-grammatical part).");
      System.err.println("<HTMLfile> - The HTML-file to be converted,");
      System.err.println("             may contain quotes (enclose the filename by quotes).");
      System.exit(63);
    }
    DeepFileSet files=new DeepFileSet(argv[0]);
    if(!files.iterator().hasNext()) {
      System.err.println("Can't find "+argv[0]);
      return;
    }
    for(File f : files) main(f);
  }
}