package leila.parsing;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FilterReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.net.URL;
import java.nio.CharBuffer;

import javatools.administrative.D;
import javatools.parsers.Char;

/** This class is part of LEILA (http://mpii.de/yago-naga/leila).
It is licensed under the Creative Commons Attribution License 
(http://creativecommons.org/licenses/by/3.0) 
by the author Fabian M. Suchanek (http://suchanek.name).
<P>
  
The HTML-Reader reads normalized characters from a HTML-file.
See Char.java for the definition of "normalized".
Tags are returned as "TAG_<i>tagname</i>".<BR>
Example:
   <PRE>
         HTMLReader r=new HTMLReader(new FileReader("index.html"));
         String s;
         while(null!=(s=r.readString())) System.out.print(s);

         -->
             This is the HTML-file, with resolved ampersand sequences
             and with all characters normalized, even the umlauts ae,
             oe, ue. Tags appear like TAG_I this TAG_/I.
   </PRE>
 */
public class HTMLReader extends FilterReader {
  /** Constructs a HTMLReader from a Reader */
  public HTMLReader(Reader s) {
    super(new BufferedReader(s));
  }

  /** Constructs a HTMLReader for an URL 
   * @throws IOException */
  public HTMLReader(URL url) throws IOException {
    this(new InputStreamReader(url.openStream()));
  }
  
  /** Constructs a HTMLReader from a File */
  public HTMLReader(File f) throws FileNotFoundException {
    super(new BufferedReader(new FileReader(f)));
  }

  /** Reads a sequence of characters
   * up to the blank following the nth char, ignores tags */
  public String readText(int n) throws IOException {
    StringBuilder l=new StringBuilder();
    String c;
    while(null!=(c=readTextChar()) && !(c.startsWith(" ") && l.length()>=n)) l.append(c);
    if(c==null && l.length()==0) return(null);
    return(l.toString());
  }

  /** Reads a character, ignores tags */
  public String readTextChar() throws IOException {
    String s=readString();
    while(s!=null && s.startsWith("TAG_")) s=readString();
    return(s);
  }

  /** Returns a single character. Deprecated! Use readString() instead,
    * because a single read character (like umlauts) may translate to
    * multiple output characters.
    * Tags are not returned */
  public int read() throws IOException {
    String s=readString();
    if(s==null) return(-1);
    if(s.startsWith("TAG_")) return(read());
    return(s.charAt(0));
  }

  protected String internalBuf="";
  
  /** Reads into a charbuffer*/
  public int read(CharBuffer buffi) throws IOException {
    if(internalBuf.length()>0) {
      int l=internalBuf.length()<buffi.remaining()?internalBuf.length():buffi.remaining();
      buffi.put(internalBuf.substring(0,l));
      internalBuf=internalBuf.substring(l);
      return(l);
    }
    String res=readString();
    if(res==null) return(-1);
    if(res.startsWith("TAG_")) res=' '+res+' ';
    int l=res.length()<buffi.remaining()?res.length():buffi.remaining();        
    buffi.put(res.substring(0,l));
    internalBuf=res.substring(l);
    return(l);
  }
  
  /** Holds the content of the last tag*/
  protected String tagContent=null;
  
  /** Returns the content of the last tag*/
  public String getLastTagContent() {
    return(tagContent);
  }
  
  /** Static variable for readString()*/
  protected boolean wasWhiteSpace=false;
  /** Reads a character, returns null for EndOfFile and
    * "TAG_tagname" for tags.
    * Any whitespace is returned as one blank. */
  public String readString() throws IOException {
    int c;
    String result;
    while(-1!=(c=in.read())) {
      switch(c) {
        case '<': StringBuilder t=new StringBuilder();
                  while((c=in.read())=='/' || c=='!' || Character.isLetterOrDigit(c)) t.append((char)c);
                  String tag=t.toString().toUpperCase();
                  t.setLength(0);
                  while(c!=-1 && c!='>') t.append((char)(c=in.read()));
                  tagContent=t.length()>0?t.toString().substring(0,t.length()-1):"";
                  if(tag.equals("SCRIPT")) {
                        while(c!=-1 && (c!='<' || in.read()!='/')) c=in.read();
                        while(c!=-1 && c!='>') c=in.read();
                        return(readString());
                  }
                  return("TAG_"+tag);
        case '&':String a="&";
                 while(!Character.isWhitespace((char)(c=in.read())) && c!=';' && c!=-1) a+=(char)c;
                 if(c!=-1) a+=(char)c;
                 c=Char.eatAmpersand(a,Char.eatLength);
                 if(Char.eatLength[0]<=0) return(a);
                 result=Char.normalize(c)+a.substring(Char.eatLength[0]);
                 if(result.length()!=0) return(result);
                 else continue;
        default: int l=Char.Utf8Length((char)c);
                 if(l==2) c=Char.eatUtf8((char)c+""+(char)in.read(), Char.eatLength);
                 if(l==3) c=Char.eatUtf8((char)c+""+(char)in.read()+(char)in.read(), Char.eatLength);
                 result=Char.normalize(c);
                 if(result.length()==0) continue;
                 if(Character.isWhitespace(result.charAt(0))) {
                   if(wasWhiteSpace) return(readString());
                   wasWhiteSpace=true;
                   return(" ");
                 }                 
                 wasWhiteSpace=false;
                 return(result);
      }
    }
    return(null);
  }

  /** Seeks the next tag of name <I>t</I> and returns all text
   * to the terminating tag /<I>t</I>. Nesting is not supported.
   * Returns null if <I>t</I> was not found. The result text is
   * normalized.
   */
  public String readTaggedText(String t) throws IOException {
    t=t.toUpperCase();
    String tagend="TAG_/"+t;    
    if(!scrollTo("TAG_"+t)) return(null);
    String c="";
    StringBuilder result=new StringBuilder();
    while(!tagend.equals(c=readString()) && c!=null) {
      if(!c.startsWith("TAG_")) result.append(c);
    }
    return(Char.decodeAndNormalize(result.toString()));
  }


  /** Seeks a specific string and scrolls to it, returns TRUE if found 
   * @throws IOException */
  public boolean scrollTo(String s) throws IOException {
    int pos=0;
    while(pos<s.length()) {
      String r=readString();
      if(r==null) return(false);
      if(s.length()>=pos+r.length() && s.substring(pos,pos+r.length()).equals(r)) pos+=r.length();
      else pos=0;
    }
    return(true);
  }
  
  /** Test routine */
  public static void main(String[] argv) throws Exception {
    D.p("Enter the name of an HTML-file and hit ENTER");
    HTMLReader r=new HTMLReader(new FileReader(D.r()));
    String s;
    while(null!=(s=r.readString())) System.out.print(s);
    r.close();
  }

}
