package leila.parsing;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.util.Map;
import java.util.TreeMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
/** This class is part of LEILA (http://mpii.de/yago-naga/leila).
It is licensed under the Creative Commons Attribution License 
(http://creativecommons.org/licenses/by/3.0) 
by the author Fabian M. Suchanek (http://suchanek.name).
<P>
  
   The NumberParser normalizes number expressions in English natural language text.
   It can work with expressions like
   <PRE>
         10 million meters
         7 inches
         2.3 sq ft
         12:30 pm
         100 km
         12 ml
         10 Mb
   </PRE>
   Example:
   <PRE>
         System.out.println(NumberParser.normalize("It was 1.2 inches long"));
         --> "It was 0.030479999999999997 meters long"
         
         System.out.println(toLong("more than ten kB"));
         --> 10000.0
   </PRE>
  */

public class NumberParser {

  /** Creates a normalized number from a number and a type.
   * In the current implementation, the normalized number is just
   * a string of the number, followed by the type in isPlural. */
  public static final String newNumber(String n,String type) {
    if(type.equals("-th")) return(n+type);
    if(!type.endsWith("s")) type+='s';
    return(n+' '+type);
  }

  /** Creates a normalized number without isPlural form */
  protected static final String newNumberNoPl(String n,String type) {
    if(type.equals("-th")) return(n+type);
    return(n+' '+type);
  }
  /** Creates a normalized number without a type.
   * @see newNumber(String n, String type)*/
  public static final String newNumber(String n) {
    return(n);
  }

  /** Creates a normalized number from a double and a type.
   * @see newNumber(String n, String type) */
  public static final String newNumber(double d,String type) {
    return(newNumber(d+"",type));
  }

  /** Two numbers match if their quotient is inside 1+-EPSILON */
  public static final double EPSILON=0.2;

  /** Maps decimal prefixes (like "giga") to their double */
  public static final Map<String,Double> prefixes=new TreeMap<String,Double>();
  static {
    prefixes.put("tera",1000000000000.0);
    prefixes.put("T",1000000000000.0);
    prefixes.put("giga",1000000000.0);
    prefixes.put("G",1000000000.0);
    prefixes.put("M",1000000.0);
    prefixes.put("mega",1000000.0);
    prefixes.put("kilo",1000.0);
    prefixes.put("k",1000.0);
    prefixes.put("deci",0.1);
    prefixes.put("d",0.1);
    prefixes.put("centi",0.01);
    prefixes.put("c",0.01);
    prefixes.put("milli",0.001);
    prefixes.put("m",0.001);
    prefixes.put("micro",0.000001);
    prefixes.put("mu",0.000001);
    prefixes.put("nano",0.000000001);
    prefixes.put("n",0.000000001);
  }

  /** A dot as a RegEx */
  private static final String DOT="\\.";
  /** A digit as a capturing RegEx */
  private static final String DIG="(\\d)";
  /** A blank as a RegEx */
  private static final String B="\\s*";
  /** A forced blank as a capturing RegEx */
  private static final String FB="(\\W)";
  /** A forced blank or "th" as a capturing RegEx */
  private static final String FBTH="(\\W|th)";
  /** A hyphen as a RegEx with blanks*/
  private static final String H=B+"(?:-+|to|until)"+B;
  /** ##th as a capturing RegEx with blank*/
  private static final String NTH="(\\d+)(?:th|rd|nd|st)"+FB;
  /** a/one/<> with a blank as a RegEx */
  private static final String A="(?:a|one|A|One)"+B;
  /** [of] with a blank as a RegEx */
  //private static final String OF=B+"(?:of)?";
  /** Prefixes as a capturing RegEx */
  private static final String P="(tera|T|giga|G|mega|M|kilo|k|deci|d|centi|c|milli|m|micro|mu|nano|n|)";
  /** A number as a capturing RegEx */
  private static final String FLOAT="(-?\\d+(?:\\.[0-9]+)?(?:[Ee]\\-?[0-9]+)?)";
  private static final String INT="([0-9]+)";
  private static final String UNIT="([a-z-]*)";
  private static final Pattern NUMBERPATTERN=Pattern.compile(FLOAT+B+UNIT);

  /** The universal unspecified normalized number */
  public static final String ANYNUMBER=newNumber("42");

  /** Tells whether this string is a normalized number */
  public static boolean isNumber(String s) {
    return(s.matches(FLOAT));
  }

  /** Tells whether this string is a normalized integer number */
  public static boolean isInt(String s) {
    return(s.matches(INT));
  }

  /** Just a pair of a Pattern and a replacement string */
  private static class FindReplace {
    public Pattern pattern;
    public String replacement;
    public String toString() {
      return(pattern+"\n--> "+replacement);
    }
    public FindReplace(String f,String r) {
      pattern=Pattern.compile(f);
      replacement=r;
    }
  }

  /** Holds the number patterns */
  private static final FindReplace[] patterns=new FindReplace[]{
  // The internal order of the patterns is essential!

    //  --------- separators ------------
    // c.##
    new FindReplace("(\\W)c\\.? ?(\\d)", "$1 about $2"),
    // #-#
    new FindReplace("(\\d)-(\\d)","$1 - $2"),
    // 1,000
    new FindReplace("(\\d+),(\\d{3}),?(\\d{3})?,?(\\d{3})?,?(\\d{3})?","$1$2$3$4$5"),
    // 1 000
    new FindReplace("(\\d{1,3}) (\\d{3}) ?(\\d{3})? ?(\\d{3})?","$1$2$3$4"),
    // .09
    new FindReplace(" "+DOT+INT," 0\\.$1"),
    //  --------- 2-12 ------------
    new FindReplace(FB+"two"+FB,"$12$2"),
    new FindReplace(FB+"three"+FB,"$13$2"),
    new FindReplace(FB+"four"+FBTH,"$14$2"),
    new FindReplace(FB+"five"+FB,"$15$2"),
    new FindReplace(FB+"six"+FBTH,"$16$2"),
    new FindReplace(FB+"seven"+FBTH,"$17$2"),
    new FindReplace(FB+"eight"+FB,"$18$2"),
    new FindReplace(FB+"nine"+FBTH,"$19$2"),
    new FindReplace(FB+"ten"+FBTH,"$110$2"),
    new FindReplace(FB+"eleven"+FBTH,"$111$2"),
    new FindReplace(FB+"twelve"+FBTH,"$112$2"),
    new FindReplace(FB+"first"+FB,"$1"+newNumber("1","-th")+"$2"),
    new FindReplace(FB+"second"+FB,"$1"+newNumber("2","-th")+"$2"),
    new FindReplace(FB+"third"+FB,"$1"+newNumber("3","-th")+"$2"),
    new FindReplace(FB+"eighth"+FB,"$1"+newNumber("8","-th")+"$2"),

    //  --------- billions and millions ------------
    // # billion
    new FindReplace(INT+DOT+DIG+B+"billion","$1$200000000"),
    new FindReplace(INT+DOT+DIG+DIG+B+"billion","$1$2$30000000"),
    new FindReplace(INT+B+"billion","$1000000000"),
    // # million
    new FindReplace(INT+DOT+DIG+B+"million","$1$200000"),
    new FindReplace(INT+DOT+DIG+DIG+B+"million","$1$2$30000"),
    new FindReplace(INT+B+"million","$1000000"),
    // # thousand
    new FindReplace(INT+DOT+DIG+B+"thousand","$1$200"),
    new FindReplace(INT+B+"thousand","$1000"),
    // # hundred
    new FindReplace(INT+B+"hundred","$100"),
    // # dozen
    new FindReplace(INT+B+"(dozens|dozen)",newNumber("$1")+"&*12&"),
    // billions
    // new FindReplace("billions"+OF,"3000000000"),
    // millions
    // new FindReplace("millions"+OF,"3000000"),
    // thousands
    // new FindReplace("thousands"+OF,newNumber("3000",CARDINAL)),
    // hundreds
    // new FindReplace("hundreds"+OF,newNumber("300",CARDINAL)),
    // dozens
    // new FindReplace("dozens"+OF,newNumber("40",CARDINAL)),
    // a billion
    new FindReplace(A+"billion","1000000000"),
    // a million
    new FindReplace(A+"million","1000000"),
    // a thousand
    new FindReplace(A+"thousand","1000"),
    // a hundred
    new FindReplace(A+"hundred","100"),
    // a dozen
    new FindReplace(A+"dozen","12"),

    // --------- Ordinal numbers -----------
    // 1st
    new FindReplace(NTH,newNumber("$1","-th")+"$2"),

    // --------- SI-Units ---------------
    new FindReplace(INT+B+"o'?clock am"+FB,newNumber("$1","oc")+"$2"),
    new FindReplace(INT+B+"o'?clock pm"+FB,newNumber("$1","oc")+"&+12&"),
    new FindReplace(INT+':'+INT+B+"am"+FB,newNumber("$1.$2","oc")+"$3"),
    new FindReplace(INT+':'+INT+B+"pm"+FB,newNumber("$1.$2","oc")+"&+12&$3"),
    new FindReplace(INT+B+"am"+FB,newNumber("$1","oc")+"$2"),
    new FindReplace(INT+B+"pm"+FB,newNumber("$1","oc")+"&+12&$2"),
    new FindReplace(FLOAT+B+'%',newNumber("$1","percent")),
    new FindReplace(FLOAT+B+"(dollars|dollar|\\$)",newNumber("$1","dollar")),
    new FindReplace("\\$"+B+FLOAT,newNumber("$1","dollar")),
    new FindReplace(FLOAT+B+P+"(metres|meters|meter|metre|m)\\^2",newNumber("$1","square-meter")+"&%$2&"),
    new FindReplace(FLOAT+B+P+"(metres|meters|meter|metre|m)\\^3",newNumber("$1","cubic-meter")+"&%$2&"),
    new FindReplace(FLOAT+B+P+"(metres|meters|meter|metre|m)"+FB,newNumber("$1","meter")+"&%$2&"+"$4"),
    new FindReplace(FLOAT+B+"(square|sq)"+B+P+"(metres|meters|meter|metre|m)",newNumber("$1","square-meter")+"&%$3&"),
    new FindReplace(FLOAT+B+"(cubic|cu)"+B+P+"(metres|meters|meter|metre|m)",newNumber("$1","cubic-meter")+"&%$3&"),
    new FindReplace(FLOAT+B+P+"(g|grams|gram)"+FB,newNumber("$1","gram")+"&%$2&$4"),
    new FindReplace(FLOAT+B+P+"(seconds|s)"+FB,newNumber("$1","second")+"&%$2&"+"$4"),
    new FindReplace(FLOAT+B+P+"(amperes|ampere|A)"+FB,newNumber("$1","ampere")+"&%$2&"+"$4"),
    new FindReplace(FLOAT+B+P+"(Kelvin|K)"+FB,newNumber("$1","kelvin")+"&%$2&"+"$4"),
    new FindReplace(FLOAT+B+P+"(Mol|mol)"+FB,newNumber("$1","mol")+"&%$2&"+"$4"),
    new FindReplace(FLOAT+B+P+"(candela|cd)"+FB,newNumber("$1","candela")+"&%$2&"+"$4"),
    new FindReplace(FLOAT+B+P+"(radians|rad)"+FB,newNumber("$1","radians")+"&%$2&"+"$4"),
    new FindReplace(FLOAT+B+P+"(hertz|Hz)"+FB,newNumber("$1","hertz")+"&%$2&"+"$4"),
    new FindReplace(FLOAT+B+P+"(newton|N)"+FB,newNumber("$1","newton")+"&%$2&"+"$4"),
    new FindReplace(FLOAT+B+P+"(joule|J)"+FB,newNumber("$1","joule")+"&%$2&"+"$4"),
    new FindReplace(FLOAT+B+P+"(watt|W)"+FB,newNumber("$1","watt")+"&%$2&"+"$4"),
    new FindReplace(FLOAT+B+P+"(pascal|Pa)"+FB,newNumber("$1","pascal")+"&%$2&"+"$4"),
    new FindReplace(FLOAT+B+P+"(lumen|lm)"+FB,newNumber("$1","lumen")+"&%$2&"+"$4"),
    new FindReplace(FLOAT+B+P+"(lux|lx)"+FB,newNumber("$1","lux")+"&%$2&"+"$4"),
    new FindReplace(FLOAT+B+P+"(coulomb|C)"+FB,newNumber("$1","coulomb")+"&%$2&"+"$4"),
    new FindReplace(FLOAT+B+P+"(volt|V)"+FB,newNumber("$1","volt")+"&%$2&"+"$4"),
    new FindReplace(FLOAT+B+P+"(ohm|O)"+FB,newNumber("$1","ohm")+"&%$2&"+"$4"),
    new FindReplace(FLOAT+B+P+"(farad|F)"+FB,newNumber("$1","farad")+"&%$2&"+"$4"),
    new FindReplace(FLOAT+B+P+"(weber|Wb)"+FB,newNumber("$1","weber")+"&%$2&"+"$4"),
    new FindReplace(FLOAT+B+P+"(tesla|T)"+FB,newNumber("$1","tesla")+"&%$2&"+"$4"),
    new FindReplace(FLOAT+B+P+"(henry|H)"+FB,newNumber("$1","henry")+"&%$2&"+"$4"),
    new FindReplace(FLOAT+B+P+"(siemens|S)"+FB,newNumber("$1","siemens")+"&%$2&"+"$4"),
    new FindReplace(FLOAT+B+P+"(becquerel|Bq)"+FB,newNumber("$1","becquerel")+"&%$2&"+"$4"),
    new FindReplace(FLOAT+B+P+"(gray|Gy)"+FB,newNumber("$1","gray")+"&%$2&"+"$4"),
    new FindReplace(FLOAT+B+P+"(sievert|Sv)"+FB,newNumber("$1","sievert")+"&%$2&"+"$4"),
    new FindReplace(FLOAT+B+P+"(katal|kat)"+FB,newNumber("$1","katal")+"&%$2&"+"$4"),
    new FindReplace(FLOAT+B+P+"(bytes|b|Bytes|B)"+FB,newNumber("$1","byte")+"&%$2&"+"$4"),

    // --------- Quasi-SI-Units ---------------
    new FindReplace(FLOAT+B+"(degrees|degree)"+B+"(Celsius|C)"+FB,newNumber("$1","kelvin")+"&+273.15&$4"),
    new FindReplace(FLOAT+B+"(minutes|minute|min)"+FB,newNumber("$1","second")+"&*60&$3"),
    new FindReplace(FLOAT+B+"(hours|hour|h)"+FB,newNumber("$1","second")+"&*3600&$3"),
    new FindReplace(FLOAT+B+"(days|day)"+FB,newNumber("$1","second")+"&*86400&$3"),
    new FindReplace(FLOAT+B+"(litres|liters|litre|liter|l|L)"+FB,newNumber("$1","cubic-meter")+"&*0.001&$3"),
    new FindReplace(FLOAT+B+"(m|milli)(litres|liters|litre|liter|l|L)"+FB,newNumber("$1","cubic-meter")+"&*0.000001&$4"), // Take care: milli must not be cubed
    new FindReplace(FLOAT+B+"(c|centi)(litres|liters|litre|liter|l|L)"+FB,newNumber("$1","cubic-meter")+"&*0.00001&$4"),
    new FindReplace(FLOAT+B+"(tonnes|tons|tonne|ton|t)"+FB,newNumber("$1","gram")+"&*1000000&$3"),

    // --------- Non-SI-Units ---------------
    new FindReplace(FLOAT+B+"nautical (miles|mile)",newNumber("$1","meter")+"&*1852&"),
    new FindReplace(FLOAT+B+"(knots|knot)",newNumber("$1","meter")+"&*1852& per hour "),
    new FindReplace(FLOAT+B+"(hectares|hectare|ha)"+FB,newNumber("$1","square-meter")+"&*10000&$3"),
    new FindReplace(FLOAT+B+"bar"+FB,newNumber("$1","pascal")+"&*100000&$2"),
    new FindReplace(FLOAT+B+"(inches|inch|in)\\^2"+FB,newNumber("$1","square-meter")+"&*0.00064516&$3"),
    new FindReplace(FLOAT+B+"(foot|ft|feet)\\^2",newNumber("$1","square-meter")+"&*0.0009290304&"),
    new FindReplace(FLOAT+B+"(miles|mile|mi)\\^2"+FB,newNumber("$1","square-meter")+"&*2589988.110336&$3"),
    new FindReplace(FLOAT+B+"(inches|inch|in)\\^3"+FB,newNumber("$1","cubic-meter")+"&*0.000016387064&$3"),
    new FindReplace(FLOAT+B+"(feet|foot|ft)\\^3",newNumber("$1","cubic-meter")+"&*0.028317&"),
    new FindReplace(FLOAT+B+"(yards|yard|yd)\\^3",newNumber("$1","cubic-meter")+"&*0.007646&"),
    new FindReplace(FLOAT+B+"(inches|inch)",newNumber("$1","meter")+"&*0.0254&"),
    new FindReplace(FLOAT+B+"(foot|feet|ft)",newNumber("$1","meter")+"&*0.3048&"),
    new FindReplace(FLOAT+B+"(yards|yard|yd)",newNumber("$1","meter")+"&*0.9144&"),
    new FindReplace(FLOAT+B+"(miles|mile|mi)"+FB,newNumber("$1","meter")+"&*1609.344&$3"),
    new FindReplace(FLOAT+B+"(square|sq)"+B+"(inches|inch|in)"+FB,newNumber("$1","square-meter")+"&*0.00064516&$4"),
    new FindReplace(FLOAT+B+"(square|sq)"+B+"(foot|ft|feet)",newNumber("$1","square-meter")+"&*0.09290304&"),
    new FindReplace(FLOAT+B+"(acres|acre)",newNumber("$1","square-meter")+"&*4046.8564224&"),
    new FindReplace(FLOAT+B+"(square|sq)"+B+"(miles|mile|mi)"+FB,newNumber("$1","square-meter")+"&*2589988.11&$4"),
    new FindReplace(FLOAT+B+"(cubic|cu)"+B+"(inches|inch|in)"+FB,newNumber("$1","cubic-meter")+"&*0.000016387064&$4"),
    new FindReplace(FLOAT+B+"(cubic|cu)"+B+"(feet|foot|ft)",newNumber("$1","cubic-meter")+"&*0.0283168466&"),
    new FindReplace(FLOAT+B+"(cubic|cu)"+B+"(yards|yard|yd)",newNumber("$1","cubic-meter")+"&*0.764554858&"),
    new FindReplace(FLOAT+B+"acre-foot",newNumber("$1","cubic-meter")+"&*12334.818&"),
    new FindReplace(FLOAT+B+"(gallon|gallons|gal)",newNumber("$1","cubic-meter")+"&*0.0037854118&"),
    new FindReplace(FLOAT+B+"(ounces|ounce|oz)",newNumber("$1","gram")+"&*28.3495231&"),
    new FindReplace(FLOAT+B+"(pounds|pound|lb)",newNumber("$1","gram")+"&*453.59237&"),
    new FindReplace(FLOAT+B+"(stones|stone)",newNumber("$1","gram")+"&*6350.29318&"),
    new FindReplace(FLOAT+B+"(degrees? Fahrenheit|degrees? F|Fahrenheit)",newNumber("$1","kelvin")+"&+459.67&&*0.55555555555&"),
    new FindReplace(FLOAT+B+"(degrees|degree)",newNumber("$1","radians")+"&*0.00555555555555&"),

    // ---------- Hyphens ---------------
    new FindReplace(FLOAT+H+FLOAT+B+UNIT+"&%([^&]*)&",newNumberNoPl("$1","$3")+"&%$4& to "+newNumberNoPl("$2","$3")+"&%$4&"),
    new FindReplace(FLOAT+H+FLOAT+B+UNIT+"&\\*([^&]*)&",newNumberNoPl("$1","$3")+"&*$4& to "+newNumberNoPl("$2","$3")+"&*$4&"),
    new FindReplace(FLOAT+H+FLOAT,newNumber("$1")+" to "+newNumber("$2"))

/*    // --------- Speed ----------
    new FindReplace(MNUMBER+" &%([^&]*)&"+B+"(per|/)"+B+"(hour|h)"+FB,newNumber("$1","ms")+"&*3600&&%$2&$5"),
    new FindReplace(MNUMBER+" &\\*([^&]*)&"+B+"(per|/)"+B+"(hour|h)"+FB,newNumber("$1","ms")+"&*3600&&*$2&$5"),
    new FindReplace(MNUMBER+B+"(per|/)"+B+"(hour|h)"+FB,newNumber("$1","ms")+"$4"),
    new FindReplace(MNUMBER+B+"(per|/)"+B+"(second|s)"+FB,newNumber("$1","ms")+"$4"),        */

  };

  /** Holds the pattern that determines whether a string contains a number */
  private static final Pattern numberPattern=Pattern.compile("two|three|four|five|six|seven|eight|nine|ten|eleven|twelve|million|billion|"+
         "hundred|thousand|first|second|third|fourth|fifth|\\d|dozen");
  /** Normalizes all numbers in a String */
  public static String normalize(String s) {
    if(!numberPattern.matcher(s).find()) return(s);
    s=' '+s+' ';
    // Find patterns
    for(FindReplace fr : patterns) {
      Matcher m=fr.pattern.matcher(s);
      if(m.find()) {
         s=m.replaceAll(fr.replacement);
      }
    }
    // Perform additions (prior to multiplications for Fahrenheit)
    Pattern p = Pattern.compile(FLOAT+B+UNIT+"&\\+(-?[0-9\\.]*)&");
    Matcher m = p.matcher(s);
    StringBuffer result1 = new StringBuffer();
    while (m.find()) {
      String type=m.group(2);
      double number=Double.parseDouble(m.group(1));
      double add=Double.parseDouble(m.group(3));
      m.appendReplacement(result1,newNumber(number+add,type));
    }
    m.appendTail(result1);
    // Perform multiplications
    p = Pattern.compile(FLOAT+B+UNIT+"&\\*([0-9\\.]*)&");
    m = p.matcher(result1);
    StringBuffer result2 = new StringBuffer();
    while (m.find()) {
      String type=m.group(2);
      double number=Double.parseDouble(m.group(1));
      double mul=Double.parseDouble(m.group(3));
      m.appendReplacement(result2,newNumber(number*mul,type));
    }
    m.appendTail(result2);
    // Perform prefixes
    p = Pattern.compile(FLOAT+B+UNIT+"&%([^&]*)&");
    m = p.matcher(result2);
    StringBuffer result3 = new StringBuffer();
    while (m.find()) {
      if(m.group(3).length()==0) {
        m.appendReplacement(result3,newNumber(m.group(1),m.group(2)));
        continue;
      }
      String type=m.group(2);
      double number=Double.parseDouble(m.group(1));
      double prefix=prefixes.get(m.group(3));
      if(Character.isDigit(type.charAt(1))) prefix=Math.pow(prefix,type.charAt(1)-48);
      m.appendReplacement(result3,newNumber(number*prefix,type));
    }
    m.appendTail(result3);
    // Perform fractions
    p = Pattern.compile(INT+" "+INT+"/"+INT);
    m = p.matcher(result3);
    StringBuffer result4 = new StringBuffer();
    while (m.find()) {
      double n1=Double.parseDouble(m.group(2));
      double n2=Double.parseDouble(m.group(3));
      m.appendReplacement(result4,newNumber(m.group(1)+(""+n1/n2).substring(1)));
    }
    m.appendTail(result4);
    return(result4.toString().trim());
  }

  /** Normalizes all numbers in a StringBuilder
    * Convenience method for the more efficient String-method */
  public static StringBuilder normalize(StringBuilder s) {
    return(new StringBuilder(normalize(s.toString())));
  }

  /** Tells whether two normalized numbers are the same (within EPSILON).
   * Any characters preceding or succeeding the numbers are ignored. */
  public static boolean match(String d1, String d2) {
    Matcher m1=NUMBERPATTERN.matcher(d1);
    if(!m1.find()) return(false);
    Matcher m2=NUMBERPATTERN.matcher(d2);
    if(!m2.find()) return(false);
    if(!m1.group(2).equals(m2.group(2))) return(false);
    if(d1.equals(ANYNUMBER) || d2.equals(ANYNUMBER)) return(true);
    double n1=Double.parseDouble(m1.group(1));
    double n2=Double.parseDouble(m2.group(1));
    if(n1==0.0) return(Math.abs(n2)<EPSILON);
    return(n1/n2>(1-EPSILON) && n1/n2<(1+EPSILON));
  }

  /** Extracts the pure number from a String containing a normalized number, else "" */
  public static String getNumber(String d) {
    Matcher m=NUMBERPATTERN.matcher(d);
    if(m.find()) return(m.group(1));
    return("");
  }

  /** Converts a String that contains a (possibly non-normalized) number to a double */
  public static double toDouble(String d) {
    return(Double.parseDouble(getNumber(normalize(d))));
  }

  /** Converts a String that contains a (possibly non-normalized) number to a long */
  public static long toLong(String d) {
    return((long)toDouble(d));
  }
  
  /** Test method*/
  public static void main(String[] argv) throws Exception {
    System.out.println("Enter a string containing a number and hit ENTER. Press CTRL+C to abort");
    while(true) {
     String in= new BufferedReader(new InputStreamReader(System.in)).readLine();
     System.out.println(normalize(in));
     System.out.println(toDouble(in));
    }
  }

}
