package leila.patterns;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;
import java.util.Vector;

import javatools.parsers.Name;
import leila.relations.Relation;

/**
This class is part of LEILA (http://mpii.de/yago-naga/leila).
It is licensed under the Creative Commons Attribution License 
(http://creativecommons.org/licenses/by/3.0) 
by the author Fabian M. Suchanek (http://suchanek.name).

<P>

LinearLinkage.java is a pseudo-linkage of a sequence of words.
*/

public class LinearLinkage extends Linkage implements Serializable {
  /** Number of tokens to be taken into account around a word */
  public static final int WIDTH=8;
  /** Links running to the right */
  public List<Link> rightLinks=new Vector<Link>(5);
  /** Links running to the left */
  protected List<Link> leftLinks=new LinkedList<Link>();
  /** Contains the delimiters */
  protected final String DELIMS=" \t\n\r\f:(){}[]@/.,<>,!#$%^&*+=|\";\\/?";
  /** Constructs a linear linkage from a string.
   * The id of a linear linkage refers only to the first 10 characters
   * of the string (to give list entries the same id). */
  public LinearLinkage(String str) {
    id=str.length()>10?str.substring(0,10).hashCode():str.hashCode();
    StringTokenizer toki=new StringTokenizer(str,DELIMS,true);
    LinkageWord previous=null;
    while(toki.hasMoreTokens()) {
      String s=toki.nextToken();
      if(Character.isWhitespace(s.charAt(0))) continue;
      // Ignore name abbreviations
      if(s.charAt(0)=='.' && previous!=null
         && previous.word.length()==1
         && Character.isUpperCase(previous.word.charAt(0))) continue;
      // Join proper names
      if(Name.isName(s)
         && previous!=null
         && previous.is(LinkageWord.Type.NAME)) {
        previous.word+='_'+s;
        continue;
      }
      // Join Dates and Numbers
      if(s.equals("@")) {
        String date;
        do {
          if(!toki.hasMoreTokens()) break;
          date=toki.nextToken();
          s+=date;
        } while(date.charAt(0)!='@');
      }
      LinkageWord next=new LinkageWord(s);
      linkageWords.add(next);
      if(previous!=null) {
        Link l=new Link(Link.Connector.ADJACENT,Link.Direction.LEFT,previous);
        next.contexts.add(l);
        leftLinks.add(0,l);
        l=new Link(Link.Connector.ADJACENT,Link.Direction.RIGHT,next);
        previous.contexts.add(l);
        rightLinks.add(l);
      }
      previous=next;
    }
  }

  /** Returns a list of LinkagePatterns for all pairs of interesting
   *  words judged as EXAMPLEs by the relation.
   *  Ignores bridges of just 1 link or too many links */
  public List<LinkagePattern> linkagePatterns(Relation r) {
    List<LinkagePattern> result=new ArrayList<LinkagePattern>(2);
    for(int i=0;i<linkageWords.size();i++) {
      LinkageWord w1=linkageWords.get(i);
      if(!w1.is(LinkageWord.INTERESTING) || !r.possibleExample(w1)) continue;
      for(int j=i-WIDTH<0?0:i-WIDTH;j<linkageWords.size() && j<=i+WIDTH;j++) {
        if(i==j || i==j-1 || j==i-1) continue;
        LinkageWord w2=linkageWords.get(j);
        if(!w2.is(LinkageWord.INTERESTING)) continue;
        if(r.assess(w1,w2)==Relation.WordPairType.EXAMPLE) {
          LinkagePattern p=new LinkagePattern(
                                linkageWords.get(i),
                                i<j?rightLinks.subList(i,j):leftLinks.subList(leftLinks.size()-i,leftLinks.size()-j),
                                linkageWords.get(j),
                                id);
          result.add(p);
          //break;
        }
      }
    }
    return(result);
  }

  /** Returns a hashCode (Negative for LinearLinkages) */
  public int hashCode() {
    return(-Math.abs(linkageWords.hashCode()));
  }

}