package leila.patterns;
import java.io.Serializable;
/**
This class is part of LEILA (http://mpii.de/yago-naga/leila).
It is licensed under the Creative Commons Attribution License 
(http://creativecommons.org/licenses/by/3.0) 
by the author Fabian M. Suchanek (http://suchanek.name).

<P>

Link.java represents a link from one LinkageWord to another.

*/
public final class Link implements Serializable  {
  /** Type of all connectors */
  public static enum Connector {
     A,AA,AF,AL,AM,AN,AZ,B,BI,BT,BW,C,CC,CO,CP,CQ,
     CX,D,DD,DG,DP,DT,E,EA,EB,EC,EE,EF,EI,EL,EN,ER,
     EZ,FL,FM,G,GN,H,I,ID,IN,J,JG,JQ,JT,K,L,LE,LI,
     M,MF,MG,MV,MX,N,ND,NF,NI,NJ,NN,NO,NR,NS,NT,NW,
     O,OD,OF,ON,OT,OX,P,PF,PP,Q,QI,R,RS,RW,S,SF,SFI,
     SI,SX,SXI,TA,TD,TH,TI,TM,TO,TQ,TS,TW,TY,U,UN,
     V,W,WN,WR,X,Y,YP,YS,Z,
     ADJACENT,UNKNOWN,HYPHEN,TITLE};
  /** The number of bits required to store a Link*/
  public static final int MAXINT=Connector.values().length;
  /** Link direction type */
  public static enum Direction {RIGHT, LEFT};

  /** Contains the connector */
  public Connector connector;
  /** Points where the link points */
  public LinkageWord destination;
  /** Link direction */
  public Direction direction;

  /** Returns a connector for a string */
  public static Connector connectorFor(String t) {
    if(t.startsWith("ID")) {
      return(Connector.ID);
    } else {
      try {
        return(Connector.valueOf(t));
      } catch(IllegalArgumentException e) {
        return(Connector.UNKNOWN);
      }
    }
  }
  /** Creates a Link from a connector (as String) , a direction and a destination*/
  public Link(String t, Direction r,LinkageWord p) {
    // The parser generates different ID-connectors
    // map all of them to Connector.ID
    connector=connectorFor(t);
    destination=p;
    direction=r;
  }
  /** Creates a Link from a connector, a direction and a destination*/
  public Link(Connector t, Direction r,LinkageWord p) {
    connector=t;
    destination=p;
    direction=r;
  }
  /** Returns a String representation (for debugging)*/
  public String toString() {
    return(direction==Direction.RIGHT?"--"+connectorDescription[connector.ordinal()]+"--> "+destination.word:
                      destination.word+" <--"+connectorDescription[connector.ordinal()]+"--");
  }
  /** Two links are equal if connector, direction and destination are equal */
  public boolean equals(Object o) {
    return(o!=null && o instanceof Link &&
           ((Link)o).direction.equals(direction) &&
           ((Link)o).destination.equals(destination) &&
           ((Link)o).connector.equals(connector));
  }
  /** Two links are weakly equal if connector and direction are equal */
  public boolean wequals(Object o) {
    return(o!=null && o instanceof Link &&
           ((Link)o).direction.equals(direction) &&
           ((Link)o).connector.equals(connector));
  }

  /** Returns a similarity measure */
  public double similarity(Link l) {
    double result=0.0;
    if(l.direction.equals(direction)) result+=0.2;
    if(connector.equals(l.connector)) result+=0.4;
    result+=destination.similarity(l.destination)*0.4;
    return(result);
  }
  /** Returns this link with an essentialized destination */
  public Link essentialize() {
    return(new Link(connector, direction, destination.essentialize()));
  }

  /** Returns a (lossy) integer representation of this link */
  public int toInt() {
    return(connector.ordinal());
  }

  /** Returns the Connector for an int */
  public static Connector unInt(int i) {
    if(i<Connector.values().length && i>=0) return(Connector.values()[i]);
    else return(Connector.UNKNOWN);
  }

  /** Returns a hashCode */
  public int hashCode() {
    return(connector.ordinal()^(direction.ordinal()-2)^destination.hashCode());
  }

  /** Contains adescription for each connector*/
  public static String[] connectorDescription=new String[]{
    "attributiveAdjective",
    "howAdjective",
    "frontedAdjective",
    "quantifier",
    "asMuch",
    "nounModifier",
    "as",
    "object",
    "idiomaticBe",
    "frontedObject",
    "what",
    "subordinateSubject",
    "conjunction",
    "openingClause",
    "quotingVerb",
    "auxiliaryForComparison",
    "comparativeAuxiliary",
    "determiner",
    "definiteDeterminer",
    "theProperName",
    "possessiveDeterminerToGerund",
    "timeDeterminer",
    "adverbForVerb",
    "adverbForAdjective",
    "adverbForBe",
    "adverbForComparison",
    "adverbForAdverb",
    "enough",
    "afterBefore",
    "else",
    "adverbForQuantity",
    "theXTheY",
    "almostAs",
    "forLong",
    "from",
    "properNoun",
    "instance",
    "howMuch",
    "infinite",
    "idiomatic",
    "inTime",
    "preposition",
    "prepositionForProperNoun",
    "prepositionForQuestionword",
    "conjunctionForTime",
    "verbParticle",
    "superlative",
    "comparison",
    "verbalPreposition",
    "postNounModifier",
    "someOfThem",
    "prepositionalNounModifier",
    "modifier",
    "commaModifier",
    "not",
    "number",
    "idiomaticNumber1",
    "idiomaticNumber2",
    "idiomaticNumber3",
    "numberWords",
    "interjection",
    "fractionalSuperlative",
    "one",
    "notTo",
    "fractionalNumber",
    "object",
    "distance",
    "verbalOf",
    "onDay",
    "timeObject",
    "fillerObject",
    "be",
    "beQuestion",
    "pastParticiple",
    "question",
    "verbalQuestion",
    "relativeClause",
    "subjectRelativeClause",
    "rightWall",
    "subject",
    "fillerSubject",
    "inverseFillerSubject",
    "subjectVerbInversion",
    "firstPersonSubject",
    "inverseFirstPersonSubject",
    "adjectiveForMonth",
    "timeOfDay",
    "that",
    "title",
    "monthDay",
    "verbTo",
    "frontedTimeObject",
    "subjunctiveThat",
    "dayDate",
    "dateNumber",
    "idiomaticNoun",
    "untilAfter",
    "idiomaticFarVerb",
    "subjectWall",
    "timeWhen",
    "where",
    "punctuation",
    "idiomaticQuantity",
    "pluralPossessiveSuffix",
    "possessiveSuffix",
    "as",
    "adjacent",
    "unknown",
    "hyphen",
    "title"
   };

}