package leila.patterns;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;

import javatools.parsers.Name;
import leila.parsing.Document;
import leila.relations.Relation;

/**
This class is part of LEILA (http://mpii.de/yago-naga/leila).
It is licensed under the Creative Commons Attribution License 
(http://creativecommons.org/licenses/by/3.0) 
by the author Fabian M. Suchanek (http://suchanek.name).

<P>

Linkage.java represents a sentence with its links. A linkage is
constructed from an LGO-file (Link Grammar Output).
*/

public class Linkage implements Serializable {
  /** Contains last sentence to determine whether to call newDocument */
  protected static int oldSentence=0;
  /** words of the sentence in sentence order */
  public List<LinkageWord> linkageWords=new ArrayList<LinkageWord>(40);
  /** Empty sentence list */
  protected List<LinkageWord> EMPTYLINKAGEWORDS=new ArrayList<LinkageWord>(0);
  /** Identifies the sequence of words */
  public int id=0;
  /** Constructs a linkage from a list of lines like<PRE>
        +---------------------Xp---------------------+
        |           +-------Ss------+                |
        +-----Wd----+----Bs---+     |                |
        |      +-Ds-+-Rn-+Sp*i+     +---Pv--+--Os-+  |
        |      |    |    |    |     |       |     |  |
    LEFT-WALL the man.n I.p saw.v was.v called.v Bob .</PRE>
  */
  public Linkage(List<String> lines, Document doc) {
    String sentence=lines.remove(lines.size()-1)+' ';
    // Skip siamese twins (bug of HTML2LGI led to concatenation of 2 sentences)
    if(sentence.indexOf("[.]")!=-1 && sentence.indexOf("[.]")<sentence.length()-6) {
      linkageWords=EMPTYLINKAGEWORDS;
      return;
    }
    int[] position2wordNum=new int[sentence.length()+10];
    try {
      // Fill linkageWords and position2wordNum, split hyphenated words
      for(int i=0; ; ) {
        int j=i;
        i=sentence.indexOf(' ',i+1);
        if(i==-1) break;
        Arrays.fill(position2wordNum,j,i,linkageWords.size());
        // Does the word contain exactly one hyphen to a non-name?
        int h=sentence.substring(j,i).lastIndexOf('-');
        if(h!=-1 && h!=0 && h!=sentence.length()
              && sentence.substring(j,i).indexOf('-')==h
              && !Name.isName(sentence.substring(j+h+1))) {
          LinkageWord tail=new LinkageWord(sentence.substring(j,j+h));
          LinkageWord head=new LinkageWord(sentence.substring(j+h+1,i));
          tail.addLink(new Link(Link.Connector.HYPHEN,Link.Direction.RIGHT,head));
          head.addLink(new Link(Link.Connector.HYPHEN,Link.Direction.LEFT,tail));
          linkageWords.add(head);
          linkageWords.add(tail); // In this order to have future links refer to head
          id^=head.word.hashCode();
        } else {
          LinkageWord linkageWord=new LinkageWord(sentence.substring(j,i));
          linkageWords.add(linkageWord);
          if(!linkageWord.word.endsWith("-WALL")) {
            // Sometimes the same word is labeled as a noun (=> stemming)
            // and as a verb (=> no stemming)
            id^=linkageWord.word.length()>3?linkageWord.word.substring(0,3).hashCode():
              linkageWord.word.hashCode();
          }
        }
        i++; // Overcome the blank
      }
      // Collect the linkages bottom-up
      for(int l=lines.size()-2; l>=0; l-- ) {
        String line=lines.get(l);
        // Collect all connectors in this line
        for(int i=line.indexOf('+'); i!=-1; i=line.indexOf('+',i)) {
          // Get left word
          int leftIndex=position2wordNum[i];
          LinkageWord leftWord=linkageWords.get(leftIndex);
          // Get connector
          i++ ;
          while(line.charAt(i)=='-') i++ ;
          String connector="";
          while(i<line.length()&&Character.isUpperCase(line.charAt(i)))
            connector+=line.charAt(i++ );
          // Get right word
          i=line.indexOf('+', i);
          int rightIndex=position2wordNum[i];
          LinkageWord rightWord=linkageWords.get(rightIndex);
          // If the right word was hyphenated, connect to tail of right word
          // to deal with "New York-based"
          if(rightWord.contexts.size()>0 && rightWord.contexts.get(0).connector==Link.Connector.HYPHEN) {
            //D.p(sentence);
            rightWord=linkageWords.get(++rightIndex);
            //D.p(leftWord);
            //D.p(rightWord);
          }
          Link.Connector c=Link.connectorFor(connector);
          // Check whether the connector is a contractor
          // Do not connect if the left word was hyphenated
          //if(isContractor(c) && !(leftWord.contexts.size()>0 && leftWord.contexts.get(0).connector==Link.Connector.HYPHEN)){
          // Use Name.isName instead of LinkageWord.is(NAME), because
          // many capital letter words are stopwords
          if(isContractor(c) && Name.isName(leftWord.word)
             && Name.isName(rightWord.word)
             && !(leftWord.contexts.size()>0
             && leftWord.contexts.get(0).connector==Link.Connector.HYPHEN)){
            // Create the new word, take over context from leftWord
            rightWord.set(leftWord.word+'_'+rightWord.word);
            for(Link ll : leftWord.contexts) {
              if(ll.destination==rightWord) continue;
              rightWord.addLink(ll);
              for(Link ll2 : ll.destination.contexts) {
                if(ll2.destination==leftWord) ll2.destination=rightWord;
              }
            }
            // If the left word is a title, keep it, else delete it
            if(Name.isTitle(leftWord.word)) {
              leftWord.contexts=new ArrayList<Link>(1);
              leftWord.addLink(new Link(Link.Connector.TITLE,Link.Direction.RIGHT,rightWord));
              rightWord.addLink(new Link(Link.Connector.TITLE,Link.Direction.LEFT,leftWord));
            } else {
             linkageWords.remove(leftIndex);
             // Adjust position2word for all future links
             for(int j=i;j<position2wordNum.length;j++) position2wordNum[j]--;
            }
          } else {
            leftWord.addLink(new Link(c,Link.Direction.RIGHT,rightWord));
            rightWord.addLink(new Link(c,Link.Direction.LEFT,leftWord));
          }
          if(i+1==line.length() || line.charAt(i+1)==' ') i++ ;
        }
      }
    }
    catch(Exception e) {
      // ArrayIndexOutOfDingsException, StringIndexException
      // ===> structure is buggy, return an empty Linkage
e.printStackTrace();
      linkageWords=EMPTYLINKAGEWORDS;
      return;
    }
    // Run through the linkagewords,
    // * resolve references
    // * add subjects as referees (in that order)
    if(oldSentence!=id) doc.newSentence();
    for(int i=0;i< linkageWords.size();i++) {
      LinkageWord l=linkageWords.get(i);
      if(l==null) continue;
      // Refresh names referred to by "his" or "her"
      if(l.word.equals("his") || l.word.equals("her")) doc.refreshTargetFor("he");
      // Resolve references
      doc.resolveReference(l);
      // Add name for completion
      doc.addName(l);
      if(l.is(LinkageWord.Type.NAME)) {
        // Try to complete the name
        doc.completeName(l);
        // If it is a company, add it as target
        if(l.is(LinkageWord.Type.COMPANY)) doc.addTarget(l);
        // If l could be a person and is a subject, add it as target
        if(l.is(LinkageWord.Type.COULDBEPERSON)) {
          for(Link link : l.contexts) {
            if(link.direction==Link.Direction.RIGHT && (
                  link.connector==Link.Connector.S
                  || link.connector==Link.Connector.MX &&
                     link.destination.word.equals("who")
                  )) {
              doc.addTarget(l);
              break;
            }
          }
        }
        continue;
      }
    }
    oldSentence=id;
  }

  /** Tells whether a connector is G, ID, MG or AN */
  private boolean isContractor(Link.Connector c) {
    return(c==Link.Connector.G || c==Link.Connector.ID ||
           c==Link.Connector.MG || c==Link.Connector.JG ||  //Worsens result because of "University of Pennsylvania"
           c==Link.Connector.AN);

  }

  /** Returns a list of LinkagePatterns for all pairs of interesting
    * words judged as EXAMPLEs by the relation */
  public List<LinkagePattern> linkagePatterns(Relation r) {
    List<LinkagePattern> result=new ArrayList<LinkagePattern>(4);
    for(int i=0;i<linkageWords.size();i++) {
      LinkageWord w1=linkageWords.get(i);
      if(!w1.is(LinkageWord.INTERESTING) || !r.possibleExample(w1)) continue;
      for(int j=0;j<linkageWords.size();j++) {
        LinkageWord w2=linkageWords.get(j);
        if(i==j || !w2.is(LinkageWord.INTERESTING) || w1.word.equalsIgnoreCase(w2.word)) continue;
        if(r.assess(w1,w2)==Relation.WordPairType.EXAMPLE) {
          LinkagePattern p=linkagePattern(w1,w2,i<j?Link.Direction.RIGHT:Link.Direction.LEFT);
          if(p!=null) result.add(p);
        }
      }
    }
    return(result);
  }

  /** Returns a LinkagePattern for a pair of basis words,
      seeks only in one direction */
  protected LinkagePattern linkagePattern(LinkageWord w, LinkageWord w2,Link.Direction d) {
    List<Link> l=find(w,w2,d);
    if(l==null || l.size()==0) return(null);
    return(new LinkagePattern(w,l,w2,id));
  }

  /** Performs a depth-first search to find a bridge from one basis word to another */
  protected List<Link> find(LinkageWord w, LinkageWord w2,Link.Direction d) {
    if(w==w2) return(new LinkedList<Link>());
    for(Link l : w.contexts) {
      if(l.direction!=d) continue;
      List<Link> f=find(l.destination,w2,d);
      if(f!=null) {
        f.add(0,l);
        return(f);
      }
    }
    return(null);
  }

  /** Returns a String representation (for debugging)*/
  public String toString() {
    StringBuilder b=new StringBuilder();
    for(LinkageWord w: linkageWords) b.append(w+" "+w.contexts+"\n");
    return(b.toString());
  }
  /** Returns a hashCode (Positive for normal Linkages) */
  public int hashCode() {
    return(Math.abs(linkageWords.hashCode()));
  }

  /** For subclasses */
  protected Linkage() {
  }

}