package leila.patterns;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;

import javatools.administrative.D;
import javatools.datatypes.FinalSet;
import javatools.parsers.Char;
import javatools.parsers.Name;
import javatools.parsers.PlingStemmer;
import leila.parsing.DateParser;
import leila.parsing.NumberParser;

/**
This class is part of LEILA (http://mpii.de/yago-naga/leila).
It is licensed under the Creative Commons Attribution License 
(http://creativecommons.org/licenses/by/3.0) 
by the author Fabian M. Suchanek (http://suchanek.name).

<P>

LinkageWord.java represents a word in a Linkage.
*/

public final class LinkageWord implements Serializable {
  /** Empty List */
  private static final List<Link>EMPTYCONTEXTS=new ArrayList<Link>(0);
  /** Dummy word */
  private static final String DUMMYWORD="@dummy@";
  /** Empty word */
  private static final String EMPTYWORD="@empty@";
  /** Type of LinkageWords */
  public enum Type {GERUND, PRONOUN, NOUN, NAME,
    ABBREVIATION, PLURAL, NUMBER, DATE, VERB,
    ADJECTIVE, UNKNOWN, STOPWORD, PERSON, COULDBEPERSON, COMPANY}
  /** Type for interesting words (NOUN|NUMBER|DATE) */
  public static final EnumSet<Type> INTERESTING=EnumSet.of(Type.NOUN,Type.NUMBER,Type.DATE);
  /** Types that become dummies when essentialized */
  public static final EnumSet<Type> DUMMIES=EnumSet.of(Type.NOUN,Type.NUMBER,Type.DATE,Type.ADJECTIVE);

  /** Name of the word */
  public String word=EMPTYWORD;
  /** Links to other words */
  public List<Link> contexts=EMPTYCONTEXTS;;
  /** Stores the type of this word */
  public EnumSet<Type> type=EnumSet.of(Type.UNKNOWN);
  /** Stores the stopword-number, if type==STOPWORD */
  public int stopWordNum=-1;

  /** Returns a hashCode */
  public int hashCode() {
    if(is(DUMMIES)) return(D.intersection(type, DUMMIES).hashCode());
    return(word.hashCode());
  }

  /** Constructs a LinkageWord from a string like "man.n" */
  public LinkageWord(String w) {
    set(w);
  }

  /** Sets the word (and types) of this LinkageWord */
  public void set(String w) {
    stopWordNum=-1;
    char pos=' ';
    word=w;
    contexts=new ArrayList<Link>(2);
    // Read the POS
    if(word.length()>2 && word.charAt(word.length()-2)=='.') {
      pos=word.charAt(word.length()-1);
      word=word.substring(0,word.length()-2);
    }
    // Normalize the word
    word=word.replaceAll("(.)\\[(\\?|!)\\]","$1");
    if(word.startsWith("[")) word=word.substring(1);
    if(word.endsWith("]")) word=word.substring(0,word.length()-1);
    word=Char.decodeAndNormalize(word);
    if(word.length()==0) word="@empty@"; // Should not happen
    // Check different cases
    if(NumberParser.isInt(word)) {
      type=EnumSet.of(Type.NUMBER,Type.DATE);
      return;
    }
    if(DateParser.isDate(word)) {
      type=EnumSet.of(Type.DATE);
      return;
    }
    if(NumberParser.isNumber(word)) {
      type=EnumSet.of(Type.NUMBER);
      return;
    }
    if(Name.isCompanyName(word)) {
      type=EnumSet.of(Type.COMPANY,Type.NAME,Type.NOUN);
      word=Name.of(word).normalize();
      return;
    }
    if(Name.isAbbreviation(word)) {
      type=EnumSet.of(Type.ABBREVIATION,Type.NAME,Type.NOUN);
      return;
    }
    if(Name.isPersonName(word)) {
      type=EnumSet.of(Type.PERSON,Type.COULDBEPERSON,Type.NAME,Type.NOUN);
      word=Name.of(word).normalize();
      return;
    }
    // Check for pronouns
    if(pronouns.indexOf(word.toLowerCase())!=-1) {
      type=EnumSet.of(Type.PRONOUN);
      word=word.toLowerCase();
      if(word.equals("they")) type.add(Type.PLURAL);
      return;
    }
    // Treat "an"
    if(word.equals("an") || word.equals("An")) word="a";
    // Check for stopwords
    stopWordNum=stopWords.indexOf(word.toLowerCase());
    if(stopWordNum!=-1) {
      type=EnumSet.of(Type.STOPWORD);
      return;
    }
    if(Name.couldBePersonName(word)) {
      type=EnumSet.of(Type.COULDBEPERSON,Type.NAME,Type.NOUN);
      return;
    }
    if(Name.isName(word)) {
      type=EnumSet.of(Type.NAME,Type.NOUN);
      return;
    }
    switch(pos) {
      case ' ':
      case 'n': type=EnumSet.of(Type.NOUN);
                String stemmed=word.substring(0,word.lastIndexOf('_')+1)+PlingStemmer.stem(word.substring(word.lastIndexOf('_')+1));
                if(!stemmed.equals(word)) type.add(Type.PLURAL);
                word=stemmed;
                return;
      case 'a': type=EnumSet.of(Type.ADJECTIVE); return;
      case 'v': type=EnumSet.of(Type.VERB); return;
      case 'p': type=EnumSet.of(Type.PRONOUN); return;
      case 'g': type=EnumSet.of(Type.GERUND); return;
    }
    type=EnumSet.of(Type.UNKNOWN);
  }

  /** Adds a link */
  public void addLink(Link l) {
    contexts.add(l);
  }

  /** Returns a String representation (for debugging)*/
  public String toString() {
    String result=word+type;
    // for(Link l : contexts) result+="\n "+l.toString();
    //return(result+'\n');
    return(result);
  }

  /** Returns TRUE if this type set contains the given type*/
  public boolean is(Type n) {
    return(type.contains(n));
  }

  /** Returns TRUE if this type set contains one of the given types*/
  public boolean is(EnumSet<Type> n) {
    return(D.containsOneOf(type,n));
  }

  /** Two LinkageWords are equal if their words are equal
    * Treats numbers and dates appropriately */
  public boolean equals(Object o) {
    if(o==null || !(o instanceof LinkageWord)) return(false);
    LinkageWord w=(LinkageWord)o;
    if(w.word.equals(DUMMYWORD)) return(D.containsOneOf(type,w.type));
    if(this.word.equals(DUMMYWORD)) return(D.containsOneOf(w.type,type));
    if(this.is(Type.DATE)) return(DateParser.match(word,w.word));
    if(this.is(Type.NUMBER)) return(NumberParser.match(word,w.word));
    return(word.equalsIgnoreCase(w.word));
  }

  /** Returns a similarity measure */
  public double similarity(LinkageWord w) {
    if(equals(w)) return(1.0);
    // We require an exact match for stopwords
    if(is(Type.STOPWORD)) {
      return(stopWordNum==w.stopWordNum?1.0:0.0);
    }
    return(D.intersection(w.type,type).size()/D.union(type,w.type).size()*0.8);
  }

  /** Returns an essentialized copy of the given LinkageWord
    * The new LinkageWord has no contexts,
    * dummies are simplified */
  public LinkageWord essentialize() {
    LinkageWord result=new LinkageWord("");
    result.contexts=EMPTYCONTEXTS;
    if(D.containsOneOf(type,DUMMIES)) {
      result.word=DUMMYWORD;
      result.type=D.intersection(type,DUMMIES);
      return(result);
    }
    result.word=this.word;
    result.type=this.type;
    result.stopWordNum=this.stopWordNum;
    return(result);
  }

  /** Maximal number of bits generated by toArray*/
  public static final int MAXBITS=4;

  /** Stores an integer representation of this word in an array,
   * starting at pos, adding offset and returning the new pos.*/
  public int toArray(int[] array,int pos,int offset) {
    for(int i=0;i<Type.values().length;i++) {  // Faster than an iterator
      if(is(Type.values()[i]) && Type.values()[i]!=Type.STOPWORD) array[pos++]=offset+i;
    }
    if(is(Type.STOPWORD)) array[pos++]=offset+Type.values().length+stopWordNum;
    return(pos);
  }

  /** Adopts the argument's word */
  public void adoptWord(LinkageWord w) {
    if(w==null) return;
    this.word=w.word;
    this.type=w.type;
    this.stopWordNum=w.stopWordNum;
  }

  /** Contains the references that will be resolved */
  public static FinalSet<String> pronouns=new FinalSet<String>(new String[]{
    "he",
    "him", // can't use "her" because "her" may be possessive pronoun
    "it",
    "she",
    "they"
  });
  /** Contains stopwords */
   public static FinalSet<String> stopWords=new FinalSet<String>(new String[]{
    "!",
    "#",
    "$",
    "%",
    "&",
    "'",
    "'s",
    "(",
    ")",
    "*",
    "+",
    ",",
    ",",
    "-",
    ".",
    "/",
    "/",
    ";",
    "<",
    "=",
    ">",
    "?",
    "@",
    "@empty@",
    "a",
    "able",
    "about",
    "above",
    "across",
    "after",
    "afterwards",
    "again",
    "against",
    "all",
    "almost",
    "alone",
    "along",
    "also",
    "although",
    "always",
    "am",
    "among",
    "amongst",
    "an",
    "and",
    "another",
    "any",
    "anybody",
    "anyhow",
    "anyone",
    "anything",
    "anyway",
    "anyways",
    "anywhere",
    "apart",
    "appear",
    "appreciate",
    "appropriate",
    "are",
    "aren't",
    "around",
    "as",
    "aside",
    "ask",
    "asking",
    "associated",
    "at",
    "available",
    "away",
    "awfully",
    "b",
    "be",
    "became",
    "because",
    "become",
    "becomes",
    "becoming",
    "been",
    "before",
    "beforehand",
    "behind",
    "being",
    "believe",
    "below",
    "beside",
    "besides",
    "best",
    "better",
    "between",
    "beyond",
    "both",
    "brief",
    "but",
    "by",
    "c",
    "c'mon",
    "c's",
    "came",
    "can",
    "can't",
    "cannot",
    "cant",
    "cause",
    "causes",
    "certain",
    "certainly",
    "changes",
    "clearly",
    "co",
    "com",
    "come",
    "comes",
    "concerning",
    "consequently",
    "consider",
    "considering",
    "contain",
    "containing",
    "contains",
    "corresponding",
    "could",
    "couldn't",
    "course",
    "currently",
    "d",
    "definitely",
    "described",
    "despite",
    "did",
    "didn't",
    "different",
    "do",
    "does",
    "doesn't",
    "doing",
    "don't",
    "done",
    "down",
    "downwards",
    "during",
    "e",
    "each",
    "edu",
    "eg",
    "eight",
    "either",
    "else",
    "elsewhere",
    "enough",
    "entirely",
    "especially",
    "et",
    "etc",
    "even",
    "ever",
    "every",
    "everybody",
    "everyone",
    "everything",
    "everywhere",
    "ex",
    "exactly",
    "example",
    "except",
    "f",
    "far",
    "few",
    "fifth",
    "first",
    "five",
    "followed",
    "following",
    "follows",
    "for",
    "former",
    "formerly",
    "forth",
    "four",
    "from",
    "further",
    "furthermore",
    "g",
    "get",
    "gets",
    "getting",
    "given",
    "gives",
    "go",
    "goes",
    "going",
    "gone",
    "got",
    "gotten",
    "greetings",
    "h",
    "had",
    "hadn't",
    "happens",
    "hardly",
    "has",
    "hasn't",
    "have",
    "haven't",
    "having",
    "he",
    "he's",
    "hello",
    "help",
    "hence",
    "her",
    "here",
    "here's",
    "hereafter",
    "hereby",
    "herein",
    "hereupon",
    "hers",
    "herself",
    "hi",
    "him",
    "himself",
    "his",
    "hither",
    "hopefully",
    "how",
    "howbeit",
    "however",
    "i",
    "i'd",
    "i'll",
    "i'm",
    "i've",
    "ie",
    "if",
    "ignored",
    "immediate",
    "in",
    "inasmuch",
    "inc",
    "indeed",
    "indicate",
    "indicated",
    "indicates",
    "inner",
    "insofar",
    "instead",
    "into",
    "inward",
    "is",
    "isn't",
    "it",
    "it'd",
    "it'll",
    "it's",
    "its",
    "itself",
    "j",
    "just",
    "k",
    "keep",
    "keeps",
    "kept",
    "know",
    "known",
    "knows",
    "l",
    "last",
    "lately",
    "later",
    "latter",
    "latterly",
    "least",
    "LEFT-WALL",
    "less",
    "lest",
    "let",
    "let's",
    "like",
    "liked",
    "likely",
    "little",
    "look",
    "looking",
    "looks",
    "ltd",
    "m",
    "mainly",
    "many",
    "may",
    "maybe",
    "me",
    "mean",
    "meanwhile",
    "merely",
    "might",
    "more",
    "moreover",
    "most",
    "mostly",
    "much",
    "must",
    "my",
    "myself",
    "n",
    "name",
    "namely",
    "nd",
    "near",
    "nearly",
    "necessary",
    "need",
    "needs",
    "neither",
    "never",
    "nevertheless",
    "new",
    "next",
    "nine",
    "no",
    "nobody",
    "non",
    "none",
    "noone",
    "nor",
    "normally",
    "not",
    "nothing",
    "novel",
    "now",
    "nowhere",
    "o",
    "obviously",
    "of",
    "off",
    "often",
    "oh",
    "ok",
    "okay",
    "old",
    "on",
    "once",
    "one",
    "ones",
    "only",
    "onto",
    "or",
    "other",
    "others",
    "otherwise",
    "ought",
    "our",
    "ours",
    "ourselves",
    "out",
    "outside",
    "over",
    "overall",
    "own",
    "p",
    "particular",
    "particularly",
    "per",
    "perhaps",
    "placed",
    "please",
    "plus",
    "possible",
    "presumably",
    "probably",
    "provides",
    "q",
    "que",
    "quite",
    "qv",
    "r",
    "rather",
    "rd",
    "re",
    "really",
    "reasonably",
    "regarding",
    "regardless",
    "regards",
    "relatively",
    "respectively",
    "s",
    "said",
    "same",
    "saw",
    "say",
    "saying",
    "says",
    "second",
    "secondly",
    "see",
    "seeing",
    "seem",
    "seemed",
    "seeming",
    "seems",
    "seen",
    "self",
    "selves",
    "sensible",
    "sent",
    "serious",
    "seriously",
    "seven",
    "several",
    "shall",
    "she",
    "should",
    "shouldn't",
    "since",
    "six",
    "so",
    "some",
    "somebody",
    "somehow",
    "someone",
    "something",
    "sometime",
    "sometimes",
    "somewhat",
    "somewhere",
    "soon",
    "sorry",
    "specified",
    "specify",
    "specifying",
    "still",
    "sub",
    "such",
    "sup",
    "sure",
    "t",
    "t's",
    "take",
    "taken",
    "tell",
    "tends",
    "th",
    "than",
    "thank",
    "thanks",
    "thanx",
    "that",
    "that's",
    "thats",
    "the",
    "their",
    "theirs",
    "them",
    "themselves",
    "then",
    "thence",
    "there",
    "there's",
    "thereafter",
    "thereby",
    "therefore",
    "therein",
    "theres",
    "thereupon",
    "these",
    "they",
    "they'd",
    "they'll",
    "they're",
    "they've",
    "think",
    "third",
    "this",
    "thorough",
    "thoroughly",
    "those",
    "though",
    "three",
    "through",
    "throughout",
    "thru",
    "thus",
    "to",
    "together",
    "too",
    "took",
    "toward",
    "towards",
    "tried",
    "tries",
    "truly",
    "try",
    "trying",
    "twice",
    "two",
    "u",
    "un",
    "under",
    "unfortunately",
    "unless",
    "unlikely",
    "until",
    "unto",
    "up",
    "upon",
    "us",
    "use",
    "used",
    "useful",
    "uses",
    "using",
    "usually",
    "uucp",
    "v",
    "value",
    "various",
    "very",
    "via",
    "viz",
    "vs",
    "w",
    "want",
    "wants",
    "was",
    "wasn't",
    "way",
    "we",
    "we'd",
    "we'll",
    "we're",
    "we've",
    "welcome",
    "well",
    "went",
    "were",
    "weren't",
    "what",
    "what's",
    "whatever",
    "when",
    "whence",
    "whenever",
    "where",
    "where's",
    "whereafter",
    "whereas",
    "whereby",
    "wherein",
    "whereupon",
    "wherever",
    "whether",
    "which",
    "while",
    "whither",
    "who",
    "who's",
    "whoever",
    "whole",
    "whom",
    "whose",
    "why",
    "will",
    "willing",
    "wish",
    "with",
    "within",
    "without",
    "won't",
    "wonder",
    "would",
    "would",
    "wouldn't",
    "x",
    "y",
    "yes",
    "yet",
    "you",
    "you'd",
    "you'll",
    "you're",
    "you've",
    "your",
    "yours",
    "yourself",
    "yourselves",
    "z",
    "zero",
    "[",
    "\"",
    "\\",
    "]",
    "^",
    "{",
    "|",
    "}"
    });
    public static final int MAXINT=stopWords.size()+Type.values().length;
}