package leila.relations;
import java.io.BufferedReader;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javatools.administrative.D;
import javatools.administrative.Parameters;
import javatools.parsers.Name;
import leila.parsing.DateParser;
import leila.parsing.NumberParser;
import leila.patterns.LinkageWord;
import leila.patterns.LinkageWord.Type;
import leila.relations.Relation.WordPairType;

/**
This class is part of LEILA (http://mpii.de/yago-naga/leila).
It is licensed under the Creative Commons Attribution License 
(http://creativecommons.org/licenses/by/3.0) 
by the author Fabian M. Suchanek (http://suchanek.name).

<P>

Birthdays.java implements the relation of persons with their birthdays.
It requires a list of birthdates in the format<P>
<PRE>
1980-04-29  Fabian M. Suchanek
...
</PRE>
The date may be given in any format accepted by the DateParser.
This class requires the following parameter in leila.ini:
<UL>
<LI> birthdaysFile = the birthdate file
</UL>
*/

public class Birthdays extends Relation {
  /** returns true */
  public boolean requiresCounterexamples() {
     return(true);
  }
  /** Holds the map of birthdays */
  public Map<String,List<String>> birthdays=new HashMap<String,List<String>>(4000);

  /** Reads the birthdays from a given file */
  public Birthdays(String f) throws Exception {
    initializeFrom(f);
  }

  /** Reads the birthdates from the given file. */
  protected void initializeFrom(String f) throws Exception {
    BufferedReader in=new BufferedReader(new FileReader(f));
    String line="a";
    while(line.length()==0 || !Character.isDigit(line.charAt(0))) line=in.readLine();
    do {
      String date=DateParser.normalize(line.substring(0,line.indexOf(' '))).trim();
      String person=new Name.PersonName(line.substring(line.indexOf(' ')+1).trim().replace(' ','_')).normalize();
      List<String> l=birthdays.get(person);
      if(l==null) birthdays.put(person,l=new ArrayList<String>(1));
      l.add(date);
    } while((line=in.readLine())!=null && line.length()!=0);
    in.close();
  }

  /** Reads the birthdays from the file given by the parameter "birthdaysFile" in leila.ini*/
  public Birthdays() throws Exception {
    Parameters.ensureParameters("birthdaysFile - a file with sample birthdates");
    initializeFrom(Parameters.get("birthdaysFile"));
  }

  /** Judges a pair of words */
  public WordPairType assess(LinkageWord w1, LinkageWord w2) {
    if(!w1.is(LinkageWord.Type.NAME) || !w2.is(LinkageWord.Type.DATE)) return(WordPairType.NOTHING);
    List<String> l=birthdays.get(w1.word);
    if(l==null) return(WordPairType.CANDIDATE);
    // We judge the correct year as TRUE if given as an INT */
    if(w2.is(LinkageWord.Type.NUMBER)) {
      String y=NumberParser.getNumber(w2.word);
      for(String s : l)
          if(DateParser.getYear(s).equals(y)) return(WordPairType.EXAMPLE);
      return(WordPairType.COUNTEREXAMPLE);
    }
    if(l.contains(w2.word)) return(WordPairType.EXAMPLE);
    return(WordPairType.COUNTEREXAMPLE);
  }

  /** tells whether a person is known in this relation */
  public boolean possibleExample(LinkageWord w){
    return(birthdays.get(w.word)!=null);
  }

  /** Test routine */
  public static void main(String[] argv) throws Exception {
    D.p("Loading...");
    Birthdays r=new Birthdays("c:\\fabian\\ontolearn\\relations\\birthdays.txt");
    D.p(" done");
    while(true) {
      LinkageWord w1=new LinkageWord(D.r());
      D.p(r.birthdays.get(w1.word));
      D.p(r.assess(w1,new LinkageWord(D.r())));
    }
  }

}