package leila.relations;
import java.io.File;
import java.util.List;

import javatools.administrative.D;
import javatools.administrative.Parameters;
import javatools.database.WordNet;
import leila.patterns.LinkageWord;
import leila.patterns.LinkageWord.Type;
import leila.relations.Relation.WordPairType;

/**
This class is part of LEILA (http://mpii.de/yago-naga/leila).
It is licensed under the Creative Commons Attribution License 
(http://creativecommons.org/licenses/by/3.0) 
by the author Fabian M. Suchanek (http://suchanek.name).

<P>

InstanceOf.java implements the isa-Relation from WordNet.
The class requires the files <TT>wn_s.pl</TT> and <TT>wn_hyp.pl</TT>.
These files can be downloaded from the <A HREF=http://wordnet.princeton.edu/>WordNet</A> site.
They are part of the Prolog version of WordNet.
A copy of the files is also provided in the distribution package of LEILA.
The following parameters are required in leila.ini:
<UL>
<LI>wordNetSFile - the file wn_s.pl</LI>
<LI>wordNetHypFile - the file wn_hyp.pl</LI>
</UL>
By default, the copies that ship with LEILA are read.
*/

public class InstanceOf extends Relation {

  /** Holds the wordnet data*/
  protected WordNet wn;

  /** returns true */
  public boolean requiresCounterexamples() {
     return(true);
  }

  /** Constructor */
  public InstanceOf() throws Exception {
    wn=new WordNet(new File(Parameters.get("wordNetSFile","Relations/InstanceOf/wn_s.pl")),
          new File(Parameters.get("wordNetHypFile","../RelationData/InstanceOf/wn_hyp.pl")),
          WordNet.WordType.NOUN,7);
  }

  /** Seeks a goalWordNet.Synset upwards from startWordNet.Synset.
   * Returns 0 if not found,
   * 1 if found, but above basic level,
   * 2 if found below or equal to basic level,
   * foundAtomic = there was an atomic term below currentWordNet.Synset.
   * Initial call: foundAtomic==false, start==current
   */
  protected int seek(WordNet.Synset start,WordNet.Synset current, List<WordNet.Synset> goals, boolean foundAtomic) {
    if(goals.contains(current)) return(foundAtomic?1:2);
    if(start!=current) foundAtomic|=current.getWords().size()>0 &&
            current.getWords().get(0).indexOf('_')==-1;
    for(WordNet.Synset hyp : current.getUps()) {
      int result=seek(start,hyp,goals,foundAtomic);
      if(result!=0) return(result);
    }
    return(0);
  }

  /** Tells whether this word is a possible example */
  public boolean possibleExample(LinkageWord w) {
    return(w.is(LinkageWord.Type.NAME) && wn.word2synsets.get(w.word)!=null);
  }

  /** Judges a pair of words */
  public WordPairType assess(LinkageWord w1, LinkageWord w2) {
    // No nouns -> counterexample
    if(!w1.is(LinkageWord.Type.NOUN) || !w2.is(LinkageWord.Type.NOUN)) return(WordPairType.COUNTEREXAMPLE);
    // Look up the words
    List<WordNet.Synset> w1synsets=wn.word2synsets.get(w1.word);
    List<WordNet.Synset> w2synsets=wn.word2synsets.get(w2.word);
    // If we did not find the second, try for lowercase version
    if(w2synsets==null) {
      w2synsets=wn.word2synsets.get(w2.word.toLowerCase());
    }
    // If we still did not find the second, try for last component
    if(w2synsets==null && w2.word.indexOf('_')!=-1) {
      w2synsets=wn.word2synsets.get(w2.word.toLowerCase().substring(w2.word.lastIndexOf('_')+1));
    }
    // Find a path from w1 to w2
    int result=0;
    if(w1synsets!=null && w2synsets!=null) {
      for(int i=0;i<w1synsets.size();i++) {
        WordNet.Synset s1=w1synsets.get(i);
        if(w2synsets.contains(s1)) continue;
        int seekval=seek(s1,s1,w2synsets,false);
        if(seekval>result) result=seekval;
        //if(seekval==2 && i==1) result=3;
      }
    }
    // EXAMPLE
    // Avoid case of United_States/Chun_Doo-hwan_regime
    if(w1.is(LinkageWord.Type.NAME) && result==2 &&
          !(w2.is(LinkageWord.Type.NAME) && wn.word2synsets.get(w2.word)==null)) {
        return(WordPairType.EXAMPLE);
    }
    // COUNTEREXAMPLE
    if(w1synsets!=null && result==0 && !w1.is(LinkageWord.Type.NAME)) {
        //w1.type.add(LinkageWord.Type.NAME);
        return(WordPairType.COUNTEREXAMPLE);
    }
    // CANDIDATE
    if(w1.is(LinkageWord.Type.NAME) && w2synsets!=null && result==0) {
        return(WordPairType.CANDIDATE);
    }
    return(WordPairType.NOTHING);
  }

  public String toString() {
    return(wn.word2synsets.toString());
  }

  /** Returs the minimal number of hops necessary from one concept to another */
  public int hops(String s1,String s2) {
    if(s1.indexOf('_')!=-1) s1=s1.substring(s1.indexOf('_')+1);
    if(s2.indexOf('_')!=-1) s2=s2.substring(s2.indexOf('_')+1);
    int min=1000;
    List<WordNet.Synset> l1=wn.synsetsFor(s1);
    if(l1==null) {
      D.p("no WordNet.Synset found for "+s1);
      return(1000);
    }
    List<WordNet.Synset> l2=wn.synsetsFor(s2);
    if(l2==null) {
      D.p("no WordNet.Synset found for "+s2);
      return(1000);
    }
    for(WordNet.Synset ss1 : l1) {
      for(WordNet.Synset ss2 : l2) {
        int n=hops(ss1,ss2);
        if(n<min) min=n;
      }
    }
    return(min);
  }

  /** Helper routine for hops */
  protected int hops(WordNet.Synset s1, WordNet.Synset s2) {
    if(s1==s2) {
      return(0);
    }
    int smallest=1000;
    for(WordNet.Synset hyp : s2.getUps()) {
      int n=hops(s1,hyp);
      if(n<smallest) smallest=n;
    }
    for(WordNet.Synset hyp : s1.getUps()) {
      int n=hops(hyp,s2);
      if(n<smallest) smallest=n;
    }
    return(smallest+1);
  }

  /** Test routine */
  public static void main(String[] argv) throws Exception {
    Parameters.init("c:\\fabian\\code\\bin\\leila.ini");
    D.p("Loading...");
    InstanceOf r=new InstanceOf();
    D.p(" done");
    while(true) {
     D.p(r.wn.word2synsets.get(D.r()));
    }
  }

}