package leila.relations;
import java.io.File;
import java.util.Collections;
import java.util.List;

import javatools.administrative.D;
import javatools.administrative.Parameters;
import javatools.database.WordNet;
import leila.patterns.LinkageWord;
import leila.patterns.LinkageWord.Type;
import leila.relations.Relation.WordPairType;


/**
This class is part of LEILA (http://mpii.de/yago-naga/leila).
It is licensed under the Creative Commons Attribution License 
(http://creativecommons.org/licenses/by/3.0) 
by the author Fabian M. Suchanek (http://suchanek.name).

<P>

Synonymy.java implements the Synonymy-Relation from WordNet.
The class requires the file <TT>wn_s.pl</TT>.
It can be downloaded from the <A HREF=http://wordnet.princeton.edu/>WordNet</A> site.
It is part of the Prolog version of WordNet.
The following parameter is required in leila.ini:
<UL>
<LI>wordNetSFile - the file wn_s.pl</LI>
</UL>
*/
public class Synonymy extends Relation {

  /** Holds the wordnet data*/
  protected WordNet wn;

  /** returns true */
  public boolean requiresCounterexamples() {
     return(true);
  }

  /** Constructor */
  public Synonymy() throws Exception {
    Parameters.ensureParameters(
          "wordNetSFile - should point to the file wn_s.pl"
    );
    // Use >= 7 senses stored per word to include the meaning
    // of "charge" as "tax" and others
    wn=new WordNet(new File(Parameters.get("wordNetSFile")),
          WordNet.WordType.NOUN,7);
  }

  /** Tells whether the word is a name */
  public boolean possibleExample(LinkageWord w) {
    return(w.is(LinkageWord.Type.NAME) && wn.word2synsets.get(w.word)!=null);
  }

  /** Judges a pair of words */
  public WordPairType assess(LinkageWord w1, LinkageWord w2) {
    /*  Assumptions:
            complete on  syn(noun,noun)
            complete on  syn(Name,Name), if both Names are known
    */
    // No nouns or equal: Ignore in Training, Validation and Application
    // Use isDigit instead of is(NUMBER), because digit/letter combinations are names
    if(!w1.is(LinkageWord.Type.NOUN) || !w2.is(LinkageWord.Type.NOUN) || w1.word.equalsIgnoreCase(w2.word)
          || Character.isDigit(w1.word.charAt(0)) || Character.isDigit(w2.word.charAt(0))) {
      return(WordPairType.NOTHING);
    }
    // Look up the words
    List<WordNet.Synset> w1synsets=wn.word2synsets.get(w1.word);
    List<WordNet.Synset> w2synsets=wn.word2synsets.get(w2.word);
    boolean knows1=w1synsets!=null;
    boolean knows2=w2synsets!=null;
    boolean syn=knows1 && knows2 && !Collections.disjoint(w1synsets,w2synsets);
    // Counterexample
    if(!syn && (knows1 && knows2 ||
                knows1 && !w1.is(LinkageWord.Type.NAME) ||
                knows2 && !w2.is(LinkageWord.Type.NAME))) return(WordPairType.COUNTEREXAMPLE);
    // Example
    if(knows1 && knows2 && syn && w1.is(LinkageWord.Type.NAME) && w2.is(LinkageWord.Type.NAME)) {
      return(WordPairType.EXAMPLE);
    }
    // Candidate
    //if(w1.is(LinkageWord.Type.NAME) && w2.is(LinkageWord.Type.NAME)) return(WordPairType.CANDIDATE);
    if(w1.is(LinkageWord.Type.NOUN) && w2.is(LinkageWord.Type.NOUN) &&
          !w1.is(LinkageWord.Type.NUMBER) && !w2.is(LinkageWord.Type.NUMBER)) return(WordPairType.CANDIDATE);
    // Else we can't tell
    return(WordPairType.NOTHING);
  }

  /** Returns all synsets */
  public String toString() {
    return(wn.word2synsets.toString());
  }

  /** Test routine */
  public static void main(String[] argv) throws Exception {
    D.p("Loading...");
    Synonymy r=new Synonymy();
    D.p(" done");
    while(true) {
      LinkageWord w1=new LinkageWord(D.r());
      D.p(r.wn.word2synsets.get(w1.word));
      D.p(r.assess(w1,new LinkageWord(D.r())));
    }
  }

}